import math

import numpy as np

from yt.units.yt_array import YTArray

prec_accum = {
    int: np.int64,
    np.int8: np.int64,
    np.int16: np.int64,
    np.int32: np.int64,
    np.int64: np.int64,
    np.uint8: np.uint64,
    np.uint16: np.uint64,
    np.uint32: np.uint64,
    np.uint64: np.uint64,
    float: np.float64,
    np.float16: np.float64,
    np.float32: np.float64,
    np.float64: np.float64,
    complex: np.complex128,
    np.complex64: np.complex128,
    np.complex128: np.complex128,
    np.dtype("int"): np.int64,
    np.dtype("int8"): np.int64,
    np.dtype("int16"): np.int64,
    np.dtype("int32"): np.int64,
    np.dtype("int64"): np.int64,
    np.dtype("uint8"): np.uint64,
    np.dtype("uint16"): np.uint64,
    np.dtype("uint32"): np.uint64,
    np.dtype("uint64"): np.uint64,
    np.dtype("float"): np.float64,
    np.dtype("float16"): np.float64,
    np.dtype("float32"): np.float64,
    np.dtype("float64"): np.float64,
    np.dtype("complex"): np.complex128,
    np.dtype("complex64"): np.complex128,
    np.dtype("complex128"): np.complex128,
}


def periodic_position(pos, ds):
    r"""Assuming periodicity, find the periodic position within the domain.

    Parameters
    ----------
    pos : array
        An array of floats.

    ds : ~yt.data_objects.static_output.Dataset
        A simulation static output.

    Examples
    --------
    >>> a = np.array([1.1, 0.5, 0.5])
    >>> data = {"Density": np.ones([32, 32, 32])}
    >>> ds = load_uniform_grid(data, [32, 32, 32], 1.0)
    >>> ppos = periodic_position(a, ds)
    >>> ppos
    array([ 0.1,  0.5,  0.5])
    """

    off = (pos - ds.domain_left_edge) % ds.domain_width
    return ds.domain_left_edge + off


def periodic_dist(a, b, period, periodicity=(True, True, True)):
    r"""Find the Euclidean periodic distance between two sets of points.

    Parameters
    ----------
    a : array or list
        Either an ndim long list of coordinates corresponding to a single point
        or an (ndim, npoints) list of coordinates for many points in space.

    b : array of list
        Either an ndim long list of coordinates corresponding to a single point
        or an (ndim, npoints) list of coordinates for many points in space.

    period : float or array or list
        If the volume is symmetrically periodic, this can be a single float,
        otherwise an array or list of floats giving the periodic size of the
        volume for each dimension.

    periodicity : An ndim-element tuple of booleans
        If an entry is true, the domain is assumed to be periodic along
        that direction.

    Examples
    --------
    >>> a = [0.1, 0.1, 0.1]
    >>> b = [0.9, 0, 9, 0.9]
    >>> period = 1.0
    >>> dist = periodic_dist(a, b, 1.0)
    >>> dist
    0.346410161514
    """
    a = np.array(a)
    b = np.array(b)
    period = np.array(period)

    if period.size == 1:
        period = np.array([period, period, period])

    if a.shape != b.shape:
        raise RuntimeError("Arrays must be the same shape.")

    if period.shape != a.shape and len(a.shape) > 1:
        n_tup = tuple(1 for i in range(a.ndim - 1))
        period = np.tile(np.reshape(period, (a.shape[0],) + n_tup), (1,) + a.shape[1:])
    elif len(a.shape) == 1:
        a = np.reshape(a, (a.shape[0],) + (1, 1))
        b = np.reshape(b, (a.shape[0],) + (1, 1))
        period = np.reshape(period, (a.shape[0],) + (1, 1))

    c = np.empty((2,) + a.shape, dtype="float64")
    c[0, :] = np.abs(a - b)

    p_directions = [i for i, p in enumerate(periodicity) if p]
    np_directions = [i for i, p in enumerate(periodicity) if not p]
    for d in p_directions:
        c[1, d, :] = period[d, :] - np.abs(a - b)[d, :]
    for d in np_directions:
        c[1, d, :] = c[0, d, :]

    d = np.amin(c, axis=0) ** 2
    r2 = d.sum(axis=0)
    if r2.size == 1:
        return np.sqrt(r2[0, 0])
    return np.sqrt(r2)


def periodic_ray(start, end, left=None, right=None):
    """
    periodic_ray(start, end, left=None, right=None)

    Break up periodic ray into non-periodic segments.
    Accepts start and end points of periodic ray as YTArrays.
    Accepts optional left and right edges of periodic volume as YTArrays.
    Returns a list of lists of coordinates, where each element of the
    top-most list is a 2-list of start coords and end coords of the
    non-periodic ray:

    [[[x0start,y0start,z0start], [x0end, y0end, z0end]],
     [[x1start,y1start,z1start], [x1end, y1end, z1end]],
     ...,]

    Parameters
    ----------
    start : array
        The starting coordinate of the ray.
    end : array
        The ending coordinate of the ray.
    left : optional, array
        The left corner of the periodic domain. If not given, an array
        of zeros with same size as the starting coordinate us used.
    right : optional, array
        The right corner of the periodic domain. If not given, an array
        of ones with same size as the starting coordinate us used.

    Examples
    --------
    >>> import yt
    >>> start = yt.YTArray([0.5, 0.5, 0.5])
    >>> end = yt.YTArray([1.25, 1.25, 1.25])
    >>> periodic_ray(start, end)
    [
        [
            YTArray([0.5, 0.5, 0.5]) (dimensionless),
            YTArray([1., 1., 1.]) (dimensionless)
        ],
        [
            YTArray([0., 0., 0.]) (dimensionless),
            YTArray([0.25, 0.25, 0.25]) (dimensionless)
        ]
     ]
    """

    if left is None:
        left = np.zeros(start.shape)
    if right is None:
        right = np.ones(start.shape)
    dim = right - left

    vector = end - start
    wall = np.zeros_like(start)
    close = np.zeros(start.shape, dtype=object)

    left_bound = vector < 0
    right_bound = vector > 0
    no_bound = vector == 0.0
    bound = vector != 0.0

    wall[left_bound] = left[left_bound]
    close[left_bound] = np.max
    wall[right_bound] = right[right_bound]
    close[right_bound] = np.min
    wall[no_bound] = np.inf
    close[no_bound] = np.min

    segments = []
    this_start = start.copy()
    this_end = end.copy()
    t = 0.0
    tolerance = 1e-6
    while t < 1.0 - tolerance:
        hit_left = (this_start <= left) & (vector < 0)
        if (hit_left).any():
            this_start[hit_left] += dim[hit_left]
            this_end[hit_left] += dim[hit_left]
        hit_right = (this_start >= right) & (vector > 0)
        if (hit_right).any():
            this_start[hit_right] -= dim[hit_right]
            this_end[hit_right] -= dim[hit_right]

        nearest = vector.unit_array * np.array(
            [close[q]([this_end[q], wall[q]]) for q in range(start.size)]
        )
        dt = ((nearest - this_start) / vector)[bound].min()
        now = this_start + vector * dt
        close_enough = np.abs(now - nearest) / np.abs(vector.max()) < 1e-10
        now[close_enough] = nearest[close_enough]
        segments.append([this_start.copy(), now.copy()])
        this_start = now.copy()
        t += dt

    return segments


def euclidean_dist(a, b):
    r"""Find the Euclidean distance between two points.

    Parameters
    ----------
    a : array or list
        Either an ndim long list of coordinates corresponding to a single point
        or an (ndim, npoints) list of coordinates for many points in space.

    b : array or list
        Either an ndim long list of coordinates corresponding to a single point
        or an (ndim, npoints) list of coordinates for many points in space.

    Examples
    --------
    >>> a = [0.1, 0.1, 0.1]
    >>> b = [0.9, 0, 9, 0.9]
    >>> period = 1.0
    >>> dist = euclidean_dist(a, b)
    >>> dist
    1.38564064606

    """
    a = np.array(a)
    b = np.array(b)
    if a.shape != b.shape:
        RuntimeError("Arrays must be the same shape.")
    c = a.copy()
    np.subtract(c, b, c)
    np.power(c, 2, c)
    c = c.sum(axis=0)
    if isinstance(c, np.ndarray):
        np.sqrt(c, c)
    else:
        # This happens if a and b only have one entry.
        c = math.sqrt(c)
    return c


def rotate_vector_3D(a, dim, angle):
    r"""Rotates the elements of an array around an axis by some angle.

    Given an array of 3D vectors a, this rotates them around a coordinate axis
    by a clockwise angle. An alternative way to think about it is the
    coordinate axes are rotated counterclockwise, which changes the directions
    of the vectors accordingly.

    Parameters
    ----------
    a : array
        An array of 3D vectors with dimension Nx3.

    dim : integer
        A integer giving the axis around which the vectors will be rotated.
        (x, y, z) = (0, 1, 2).

    angle : float
        The angle in radians through which the vectors will be rotated
        clockwise.

    Examples
    --------
    >>> a = np.array([[1, 1, 0], [1, 0, 1], [0, 1, 1], [1, 1, 1], [3, 4, 5]])
    >>> b = rotate_vector_3D(a, 2, np.pi / 2)
    >>> print(b)
    [[  1.00000000e+00  -1.00000000e+00   0.00000000e+00]
    [  6.12323400e-17  -1.00000000e+00   1.00000000e+00]
    [  1.00000000e+00   6.12323400e-17   1.00000000e+00]
    [  1.00000000e+00  -1.00000000e+00   1.00000000e+00]
    [  4.00000000e+00  -3.00000000e+00   5.00000000e+00]]

    """
    mod = False
    if len(a.shape) == 1:
        mod = True
        a = np.array([a])
    if a.shape[1] != 3:
        raise ValueError("The second dimension of the array a must be == 3!")
    if dim == 0:
        R = np.array(
            [
                [1, 0, 0],
                [0, np.cos(angle), np.sin(angle)],
                [0, -np.sin(angle), np.cos(angle)],
            ]
        )
    elif dim == 1:
        R = np.array(
            [
                [np.cos(angle), 0, -np.sin(angle)],
                [0, 1, 0],
                [np.sin(angle), 0, np.cos(angle)],
            ]
        )
    elif dim == 2:
        R = np.array(
            [
                [np.cos(angle), np.sin(angle), 0],
                [-np.sin(angle), np.cos(angle), 0],
                [0, 0, 1],
            ]
        )
    else:
        raise ValueError("dim must be 0, 1, or 2!")
    if mod:
        return np.dot(R, a.T).T[0]
    else:
        return np.dot(R, a.T).T


def modify_reference_frame(CoM, L, P=None, V=None):
    r"""Rotates and translates data into a new reference frame to make
    calculations easier.

    This is primarily useful for calculations of halo data.
    The data is translated into the center of mass frame.
    Next, it is rotated such that the angular momentum vector for the data
    is aligned with the z-axis. Put another way, if one calculates the angular
    momentum vector on the data that comes out of this function, it will
    always be along the positive z-axis.
    If the center of mass is re-calculated, it will be at the origin.

    Parameters
    ----------
    CoM : array
        The center of mass in 3D.

    L : array
        The angular momentum vector.

    Optional
    --------

    P : array
        The positions of the data to be modified (i.e. particle or grid cell
        positions). The array should be Nx3.

    V : array
        The velocities of the data to be modified (i.e. particle or grid cell
        velocities). The array should be Nx3.

    Returns
    -------
    L : array
        The angular momentum vector equal to [0, 0, 1] modulo machine error.

    P : array
        The modified positional data. Only returned if P is not None

    V : array
        The modified velocity data. Only returned if V is not None

    Examples
    --------
    >>> CoM = np.array([0.5, 0.5, 0.5])
    >>> L = np.array([1, 0, 0])
    >>> P = np.array([[1, 0.5, 0.5], [0, 0.5, 0.5], [0.5, 0.5, 0.5], [0, 0, 0]])
    >>> V = p.copy()
    >>> LL, PP, VV = modify_reference_frame(CoM, L, P, V)
    >>> LL
    array([  6.12323400e-17,   0.00000000e+00,   1.00000000e+00])
    >>> PP
    array([[  3.06161700e-17,   0.00000000e+00,   5.00000000e-01],
           [ -3.06161700e-17,   0.00000000e+00,  -5.00000000e-01],
           [  0.00000000e+00,   0.00000000e+00,   0.00000000e+00],
           [  5.00000000e-01,  -5.00000000e-01,  -5.00000000e-01]])
    >>> VV
    array([[ -5.00000000e-01,   5.00000000e-01,   1.00000000e+00],
           [ -5.00000000e-01,   5.00000000e-01,   3.06161700e-17],
           [ -5.00000000e-01,   5.00000000e-01,   5.00000000e-01],
           [  0.00000000e+00,   0.00000000e+00,   0.00000000e+00]])

    """
    # First translate the positions to center of mass reference frame.
    if P is not None:
        P = P - CoM

    # is the L vector pointing in the Z direction?
    if np.inner(L[:2], L[:2]) == 0.0:
        # the reason why we need to explicitly check for the above
        # is that formula is used in denominator
        # this just checks if we need to flip the z axis or not
        if L[2] < 0.0:
            # this is just a simple flip in direction of the z axis
            if P is not None:
                P = -P
            if V is not None:
                V = -V

        # return the values
        if V is None and P is not None:
            return L, P
        elif P is None and V is not None:
            return L, V
        else:
            return L, P, V

    # Normal vector is not aligned with simulation Z axis
    # Therefore we are going to have to apply a rotation
    # Now find the angle between modified L and the x-axis.
    LL = L.copy()
    LL[2] = 0.0
    theta = np.arccos(np.inner(LL, [1.0, 0.0, 0.0]) / np.inner(LL, LL) ** 0.5)
    if L[1] < 0.0:
        theta = -theta
    # Now rotate all the position, velocity, and L vectors by this much around
    # the z axis.
    if P is not None:
        P = rotate_vector_3D(P, 2, theta)
    if V is not None:
        V = rotate_vector_3D(V, 2, theta)
    L = rotate_vector_3D(L, 2, theta)
    # Now find the angle between L and the z-axis.
    theta = np.arccos(np.inner(L, [0.0, 0.0, 1.0]) / np.inner(L, L) ** 0.5)
    # This time we rotate around the y axis.
    if P is not None:
        P = rotate_vector_3D(P, 1, theta)
    if V is not None:
        V = rotate_vector_3D(V, 1, theta)
    L = rotate_vector_3D(L, 1, theta)

    # return the values
    if V is None and P is not None:
        return L, P
    elif P is None and V is not None:
        return L, V
    else:
        return L, P, V


def compute_rotational_velocity(CoM, L, P, V):
    r"""Computes the rotational velocity for some data around an axis.

    This is primarily for halo computations.
    Given some data, this computes the circular rotational velocity of each
    point (particle) in reference to the axis defined by the angular momentum
    vector.
    This is accomplished by converting the reference frame of the center of
    mass of the halo.

    Parameters
    ----------
    CoM : array
        The center of mass in 3D.

    L : array
        The angular momentum vector.

    P : array
        The positions of the data to be modified (i.e. particle or grid cell
        positions). The array should be Nx3.

    V : array
        The velocities of the data to be modified (i.e. particle or grid cell
        velocities). The array should be Nx3.

    Returns
    -------
    v : array
        An array N elements long that gives the circular rotational velocity
        for each datum (particle).

    Examples
    --------
    >>> CoM = np.array([0, 0, 0])
    >>> L = np.array([0, 0, 1])
    >>> P = np.array([[1, 0, 0], [1, 1, 1], [0, 0, 1], [1, 1, 0]])
    >>> V = np.array([[0, 1, 10], [-1, -1, -1], [1, 1, 1], [1, -1, -1]])
    >>> circV = compute_rotational_velocity(CoM, L, P, V)
    >>> circV
    array([ 1.        ,  0.        ,  0.        ,  1.41421356])

    """
    # First we translate into the simple coordinates.
    L, P, V = modify_reference_frame(CoM, L, P, V)
    # Find the vector in the plane of the galaxy for each position point
    # that is perpendicular to the radial vector.
    radperp = np.cross([0, 0, 1], P)
    # Find the component of the velocity along the radperp vector.
    # Unf., I don't think there's a better way to do this.
    res = np.empty(V.shape[0], dtype="float64")
    for i, rp in enumerate(radperp):
        temp = np.dot(rp, V[i]) / np.dot(rp, rp) * rp
        res[i] = np.dot(temp, temp) ** 0.5
    return res


def compute_parallel_velocity(CoM, L, P, V):
    r"""Computes the parallel velocity for some data around an axis.

    This is primarily for halo computations.
    Given some data, this computes the velocity component along the angular
    momentum vector.
    This is accomplished by converting the reference frame of the center of
    mass of the halo.

    Parameters
    ----------
    CoM : array
        The center of mass in 3D.

    L : array
        The angular momentum vector.

    P : array
        The positions of the data to be modified (i.e. particle or grid cell
        positions). The array should be Nx3.

    V : array
        The velocities of the data to be modified (i.e. particle or grid cell
        velocities). The array should be Nx3.

    Returns
    -------
    v : array
        An array N elements long that gives the parallel velocity for
        each datum (particle).

    Examples
    --------
    >>> CoM = np.array([0, 0, 0])
    >>> L = np.array([0, 0, 1])
    >>> P = np.array([[1, 0, 0], [1, 1, 1], [0, 0, 1], [1, 1, 0]])
    >>> V = np.array([[0, 1, 10], [-1, -1, -1], [1, 1, 1], [1, -1, -1]])
    >>> paraV = compute_parallel_velocity(CoM, L, P, V)
    >>> paraV
    array([10, -1,  1, -1])

    """
    # First we translate into the simple coordinates.
    L, P, V = modify_reference_frame(CoM, L, P, V)
    # And return just the z-axis velocities.
    return V[:, 2]


def compute_radial_velocity(CoM, L, P, V):
    r"""Computes the radial velocity for some data around an axis.

    This is primarily for halo computations.
    Given some data, this computes the radial velocity component for the data.
    This is accomplished by converting the reference frame of the center of
    mass of the halo.

    Parameters
    ----------
    CoM : array
        The center of mass in 3D.

    L : array
        The angular momentum vector.

    P : array
        The positions of the data to be modified (i.e. particle or grid cell
        positions). The array should be Nx3.

    V : array
        The velocities of the data to be modified (i.e. particle or grid cell
        velocities). The array should be Nx3.

    Returns
    -------
    v : array
        An array N elements long that gives the radial velocity for
        each datum (particle).

    Examples
    --------
    >>> CoM = np.array([0, 0, 0])
    >>> L = np.array([0, 0, 1])
    >>> P = np.array([[1, 0, 0], [1, 1, 1], [0, 0, 1], [1, 1, 0]])
    >>> V = np.array([[0, 1, 10], [-1, -1, -1], [1, 1, 1], [1, -1, -1]])
    >>> radV = compute_radial_velocity(CoM, L, P, V)
    >>> radV
    array([ 1.        ,  1.41421356 ,  0.        ,  0.])

    """
    # First we translate into the simple coordinates.
    L, P, V = modify_reference_frame(CoM, L, P, V)
    # We find the tangential velocity by dotting the velocity vector
    # with the cylindrical radial vector for this point.
    # Unf., I don't think there's a better way to do this.
    P[:, 2] = 0
    res = np.empty(V.shape[0], dtype="float64")
    for i, rad in enumerate(P):
        temp = np.dot(rad, V[i]) / np.dot(rad, rad) * rad
        res[i] = np.dot(temp, temp) ** 0.5
    return res


def compute_cylindrical_radius(CoM, L, P, V):
    r"""Compute the radius for some data around an axis in cylindrical
    coordinates.

    This is primarily for halo computations.
    Given some data, this computes the cylindrical radius for each point.
    This is accomplished by converting the reference frame of the center of
    mass of the halo.

    Parameters
    ----------
    CoM : array
        The center of mass in 3D.

    L : array
        The angular momentum vector.

    P : array
        The positions of the data to be modified (i.e. particle or grid cell
        positions). The array should be Nx3.

    V : array
        The velocities of the data to be modified (i.e. particle or grid cell
        velocities). The array should be Nx3.

    Returns
    -------
    cyl_r : array
        An array N elements long that gives the radial velocity for
        each datum (particle).

    Examples
    --------
    >>> CoM = np.array([0, 0, 0])
    >>> L = np.array([0, 0, 1])
    >>> P = np.array([[1, 0, 0], [1, 1, 1], [0, 0, 1], [1, 1, 0]])
    >>> V = np.array([[0, 1, 10], [-1, -1, -1], [1, 1, 1], [1, -1, -1]])
    >>> cyl_r = compute_cylindrical_radius(CoM, L, P, V)
    >>> cyl_r
    array([ 1.        ,  1.41421356,  0.        ,  1.41421356])

    """
    # First we translate into the simple coordinates.
    L, P, V = modify_reference_frame(CoM, L, P, V)
    # Demote all the positions to the z=0 plane, which makes the distance
    # calculation very easy.
    P[:, 2] = 0
    return np.sqrt((P * P).sum(axis=1))


def ortho_find(vec1):
    r"""Find two complementary orthonormal vectors to a given vector.

    For any given non-zero vector, there are infinite pairs of vectors
    orthonormal to it.  This function gives you one arbitrary pair from
    that set along with the normalized version of the original vector.

    Parameters
    ----------
    vec1 : array_like
           An array or list to represent a 3-vector.

    Returns
    -------
    vec1 : array
           The original 3-vector array after having been normalized.

    vec2 : array
           A 3-vector array which is orthonormal to vec1.

    vec3 : array
           A 3-vector array which is orthonormal to vec1 and vec2.

    Raises
    ------
    ValueError
           If input vector is the zero vector.

    Notes
    -----
    Our initial vector is `vec1` which consists of 3 components: `x1`, `y1`,
    and `z1`.  ortho_find determines a vector, `vec2`, which is orthonormal
    to `vec1` by finding a vector which has a zero-value dot-product with
    `vec1`.

    .. math::

       vec1 \cdot vec2 = x_1 x_2 + y_1 y_2 + z_1 z_2 = 0

    As a starting point, we arbitrarily choose `vec2` to have `x2` = 1,
    `y2` = 0:

    .. math::

       vec1 \cdot vec2 = x_1 + (z_1 z_2) = 0

       \rightarrow z_2 = -(x_1 / z_1)

    Of course, this will fail if `z1` = 0, in which case, let's say use
    `z2` = 1 and `x2` = 0:

    .. math::

       \rightarrow y_2 = -(z_1 / y_1)

    Similarly, if `y1` = 0, this case will fail, in which case we use
    `y2` = 1 and `z2` = 0:

    .. math::

       \rightarrow x_2 = -(y_1 / x_1)

    Since we don't allow `vec1` to be zero, all cases are accounted for.

    Producing `vec3`, the complementary orthonormal vector to `vec1` and `vec2`
    is accomplished by simply taking the cross product of `vec1` and `vec2`.

    Examples
    --------
    >>> a = [1.0, 2.0, 3.0]
    >>> a, b, c = ortho_find(a)
    >>> a
    array([ 0.26726124,  0.53452248,  0.80178373])
    >>> b
    array([ 0.9486833 ,  0.        , -0.31622777])
    >>> c
    array([-0.16903085,  0.84515425, -0.50709255])
    """
    vec1 = np.array(vec1, dtype=np.float64)
    # Normalize
    norm = np.sqrt(np.vdot(vec1, vec1))
    if norm == 0:
        raise ValueError("Zero vector used as input.")
    vec1 /= norm
    x1 = vec1[0]
    y1 = vec1[1]
    z1 = vec1[2]
    if z1 != 0:
        x2 = 1.0
        y2 = 0.0
        z2 = -(x1 / z1)
        norm2 = (1.0 + z2**2.0) ** (0.5)
    elif y1 != 0:
        x2 = 0.0
        z2 = 1.0
        y2 = -(z1 / y1)
        norm2 = (1.0 + y2**2.0) ** (0.5)
    else:
        y2 = 1.0
        z2 = 0.0
        x2 = -(y1 / x1)
        norm2 = (1.0 + z2**2.0) ** (0.5)
    vec2 = np.array([x2, y2, z2])
    vec2 /= norm2
    vec3 = np.cross(vec1, vec2)
    return vec1, vec2, vec3


def quartiles(a, axis=None, out=None, overwrite_input=False):
    """
    Compute the quartile values (25% and 75%) along the specified axis
    in the same way that the numpy.median calculates the median (50%) value
    alone a specified axis.  Check numpy.median for details, as it is
    virtually the same algorithm.

    Returns an array of the quartiles of the array elements [lower quartile,
    upper quartile].

    Parameters
    ----------
    a : array_like
        Input array or object that can be converted to an array.
    axis : {None, int}, optional
        Axis along which the quartiles are computed. The default (axis=None)
        is to compute the quartiles along a flattened version of the array.
    out : ndarray, optional
        Alternative output array in which to place the result. It must
        have the same shape and buffer length as the expected output,
        but the type (of the output) will be cast if necessary.
    overwrite_input : {False, True}, optional
       If True, then allow use of memory of input array (a) for
       calculations. The input array will be modified by the call to
       quartiles. This will save memory when you do not need to preserve
       the contents of the input array. Treat the input as undefined,
       but it will probably be fully or partially sorted. Default is
       False. Note that, if `overwrite_input` is True and the input
       is not already an ndarray, an error will be raised.

    Returns
    -------
    quartiles : ndarray
        A new 2D array holding the result (unless `out` is specified, in
        which case that array is returned instead).  If the input contains
        integers, or floats of smaller precision than 64, then the output
        data-type is float64.  Otherwise, the output data-type is the same
        as that of the input.

    See Also
    --------
    numpy.median, numpy.mean, numpy.percentile

    Notes
    -----
    Given a vector V of length N, the quartiles of V are the 25% and 75% values
    of a sorted copy of V, ``V_sorted`` - i.e., ``V_sorted[(N-1)/4]`` and
    ``3*V_sorted[(N-1)/4]``, when N is odd.  When N is even, it is the average
    of the two values bounding these values of ``V_sorted``.

    Examples
    --------
    >>> a = np.arange(100).reshape(10, 10)
    >>> a
    array([[ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9],
           [10, 11, 12, 13, 14, 15, 16, 17, 18, 19],
           [20, 21, 22, 23, 24, 25, 26, 27, 28, 29],
           [30, 31, 32, 33, 34, 35, 36, 37, 38, 39],
           [40, 41, 42, 43, 44, 45, 46, 47, 48, 49],
           [50, 51, 52, 53, 54, 55, 56, 57, 58, 59],
           [60, 61, 62, 63, 64, 65, 66, 67, 68, 69],
           [70, 71, 72, 73, 74, 75, 76, 77, 78, 79],
           [80, 81, 82, 83, 84, 85, 86, 87, 88, 89],
           [90, 91, 92, 93, 94, 95, 96, 97, 98, 99]])
    >>> mu.quartiles(a)
    array([ 24.5,  74.5])
    >>> mu.quartiles(a, axis=0)
    array([[ 15.,  16.,  17.,  18.,  19.,  20.,  21.,  22.,  23.,  24.],
           [ 65.,  66.,  67.,  68.,  69.,  70.,  71.,  72.,  73.,  74.]])
    >>> mu.quartiles(a, axis=1)
    array([[  1.5,  11.5,  21.5,  31.5,  41.5,  51.5,  61.5,  71.5,  81.5,
             91.5],
           [  6.5,  16.5,  26.5,  36.5,  46.5,  56.5,  66.5,  76.5,  86.5,
             96.5]])
    """
    if overwrite_input:
        if axis is None:
            a_sorted = sorted(a.ravel())
        else:
            a.sort(axis=axis)
            a_sorted = a
    else:
        a_sorted = np.sort(a, axis=axis)
    if axis is None:
        axis = 0
    indexer = [slice(None)] * a_sorted.ndim
    indices = [int(a_sorted.shape[axis] / 4), int(a_sorted.shape[axis] * 0.75)]
    result = []
    for index in indices:
        if a_sorted.shape[axis] % 2 == 1:
            # index with slice to allow mean (below) to work
            indexer[axis] = slice(index, index + 1)
        else:
            indexer[axis] = slice(index - 1, index + 1)
        # special cases for small arrays
        if a_sorted.shape[axis] == 2:
            # index with slice to allow mean (below) to work
            indexer[axis] = slice(index, index + 1)
        # Use mean in odd and even case to coerce data type
        # and check, use out array.
        result.append(np.mean(a_sorted[tuple(indexer)], axis=axis, out=out))
    return np.array(result)


def get_perspective_matrix(fovy, aspect, z_near, z_far):
    """
    Given a field of view in radians, an aspect ratio, and a near
    and far plane distance, this routine computes the transformation matrix
    corresponding to perspective projection using homogeneous coordinates.

    Parameters
    ----------
    fovy : scalar
        The angle in degrees of the field of view.

    aspect : scalar
        The aspect ratio of width / height for the projection.

    z_near : scalar
        The distance of the near plane from the camera.

    z_far : scalar
        The distance of the far plane from the camera.

    Returns
    -------
    persp_matrix : ndarray
        A new 4x4 2D array. Represents a perspective transformation
        in homogeneous coordinates. Note that this matrix does not
        actually perform the projection. After multiplying a 4D
        vector of the form (x_0, y_0, z_0, 1.0), the point will be
        transformed to some (x_1, y_1, z_1, w). The final projection
        is applied by performing a divide by w, that is
        (x_1/w, y_1/w, z_1/w, w/w). The matrix uses a row-major
        ordering, rather than the column major ordering typically
        used by OpenGL.

    Notes
    -----
    The usage of 4D homogeneous coordinates is for OpenGL and GPU
    hardware that automatically performs the divide by w operation.
    See the following for more details about the OpenGL perspective matrices.

    https://www.tomdalling.com/blog/modern-opengl/explaining-homogenous-coordinates-and-projective-geometry/
    http://www.songho.ca/opengl/gl_projectionmatrix.html

    """

    tan_half_fovy = np.tan(np.radians(fovy) / 2)

    result = np.zeros((4, 4), dtype="float32", order="C")
    # result[0][0] = 1 / (aspect * tan_half_fovy)
    # result[1][1] = 1 / tan_half_fovy
    # result[2][2] = - (z_far + z_near) / (z_far - z_near)
    # result[3][2] = -1
    # result[2][3] = -(2 * z_far * z_near) / (z_far - z_near)

    f = z_far
    n = z_near

    t = tan_half_fovy * n
    b = -t * aspect
    r = t * aspect
    l = -t * aspect

    result[0][0] = (2 * n) / (r - l)
    result[2][0] = (r + l) / (r - l)
    result[1][1] = (2 * n) / (t - b)
    result[1][2] = (t + b) / (t - b)
    result[2][2] = -(f + n) / (f - n)
    result[2][3] = -2 * f * n / (f - n)
    result[3][2] = -1

    return result


def get_orthographic_matrix(maxr, aspect, z_near, z_far):
    """
    Given a field of view in radians, an aspect ratio, and a near
    and far plane distance, this routine computes the transformation matrix
    corresponding to perspective projection using homogeneous coordinates.

    Parameters
    ----------
    maxr : scalar
        should be ``max(|x|, |y|)``

    aspect : scalar
        The aspect ratio of width / height for the projection.

    z_near : scalar
        The distance of the near plane from the camera.

    z_far : scalar
        The distance of the far plane from the camera.

    Returns
    -------
    persp_matrix : ndarray
        A new 4x4 2D array. Represents a perspective transformation
        in homogeneous coordinates. Note that this matrix does not
        actually perform the projection. After multiplying a 4D
        vector of the form (x_0, y_0, z_0, 1.0), the point will be
        transformed to some (x_1, y_1, z_1, w). The final projection
        is applied by performing a divide by w, that is
        (x_1/w, y_1/w, z_1/w, w/w). The matrix uses a row-major
        ordering, rather than the column major ordering typically
        used by OpenGL.

    Notes
    -----
    The usage of 4D homogeneous coordinates is for OpenGL and GPU
    hardware that automatically performs the divide by w operation.
    See the following for more details about the OpenGL perspective matrices.

    http://www.scratchapixel.com/lessons/3d-basic-rendering/perspective-and-orthographic-projection-matrix/orthographic-projection-matrix
    https://www.tomdalling.com/blog/modern-opengl/explaining-homogenous-coordinates-and-projective-geometry/
    http://www.songho.ca/opengl/gl_projectionmatrix.html

    """

    r = maxr * aspect
    t = maxr
    l = -r
    b = -t

    result = np.zeros((4, 4), dtype="float32", order="C")
    result[0][0] = 2.0 / (r - l)
    result[1][1] = 2.0 / (t - b)
    result[2][2] = -2.0 / (z_far - z_near)
    result[3][3] = 1

    result[3][0] = -(r + l) / (r - l)
    result[3][1] = -(t + b) / (t - b)
    result[3][2] = -(z_far + z_near) / (z_far - z_near)

    return result


def get_lookat_matrix(eye, center, up):
    """
    Given the position of a camera, the point it is looking at, and
    an up-direction. Computes the lookat matrix that moves all vectors
    such that the camera is at the origin of the coordinate system,
    looking down the z-axis.

    Parameters
    ----------
    eye : array_like
        The position of the camera. Must be 3D.

    center : array_like
        The location that the camera is looking at. Must be 3D.

    up : array_like
        The direction that is considered up for the camera. Must be
        3D.

    Returns
    -------
    lookat_matrix : ndarray
        A new 4x4 2D array in homogeneous coordinates. This matrix
        moves all vectors in the same way required to move the camera
        to the origin of the coordinate system, with it pointing down
        the negative z-axis.

    """

    eye = np.array(eye)
    center = np.array(center)
    up = np.array(up)

    f = (center - eye) / np.linalg.norm(center - eye)
    s = np.cross(f, up) / np.linalg.norm(np.cross(f, up))
    u = np.cross(s, f)

    result = np.zeros((4, 4), dtype="float32", order="C")

    result[0][0] = s[0]
    result[0][1] = s[1]
    result[0][2] = s[2]
    result[1][0] = u[0]
    result[1][1] = u[1]
    result[1][2] = u[2]
    result[2][0] = -f[0]
    result[2][1] = -f[1]
    result[2][2] = -f[2]
    result[0][3] = -np.dot(s, eye)
    result[1][3] = -np.dot(u, eye)
    result[2][3] = np.dot(f, eye)
    result[3][3] = 1.0
    return result


def get_translate_matrix(dx, dy, dz):
    """
    Given a movement amount for each coordinate, creates a translation
    matrix that moves the vector by each amount.

    Parameters
    ----------
    dx : scalar
        A translation amount for the x-coordinate

    dy : scalar
        A translation amount for the y-coordinate

    dz : scalar
        A translation amount for the z-coordinate

    Returns
    -------
    trans_matrix : ndarray
        A new 4x4 2D array. Represents a translation by dx, dy
        and dz in each coordinate respectively.
    """
    result = np.zeros((4, 4), dtype="float32", order="C")

    result[0][0] = 1.0
    result[1][1] = 1.0
    result[2][2] = 1.0
    result[3][3] = 1.0

    result[0][3] = dx
    result[1][3] = dy
    result[2][3] = dz

    return result


def get_scale_matrix(dx, dy, dz):
    """
    Given a scaling factor for each coordinate, returns a matrix that
    corresponds to the given scaling amounts.

    Parameters
    ----------
    dx : scalar
        A scaling factor for the x-coordinate.

    dy : scalar
        A scaling factor for the y-coordinate.

    dz : scalar
        A scaling factor for the z-coordinate.

    Returns
    -------
    scale_matrix : ndarray
        A new 4x4 2D array. Represents a scaling by dx, dy, and dz
        in each coordinate respectively.
    """
    result = np.zeros((4, 4), dtype="float32", order="C")

    result[0][0] = dx
    result[1][1] = dy
    result[2][2] = dz
    result[3][3] = 1

    return result


def get_rotation_matrix(theta, rot_vector):
    """
    Given an angle theta and a 3D vector rot_vector, this routine
    computes the rotation matrix corresponding to rotating theta
    radians about rot_vector.

    Parameters
    ----------
    theta : scalar
        The angle in radians.

    rot_vector : array_like
        The axis of rotation.  Must be 3D.

    Returns
    -------
    rot_matrix : ndarray
         A new 3x3 2D array.  This is the representation of a
         rotation of theta radians about rot_vector in the simulation
         box coordinate frame

    See Also
    --------
    ortho_find

    Examples
    --------
    >>> a = [0, 1, 0]
    >>> theta = 0.785398163  # pi/4
    >>> rot = mu.get_rotation_matrix(theta, a)
    >>> rot
    array([[ 0.70710678,  0.        ,  0.70710678],
           [ 0.        ,  1.        ,  0.        ],
           [-0.70710678,  0.        ,  0.70710678]])
    >>> np.dot(rot, a)
    array([ 0.,  1.,  0.])
    # since a is an eigenvector by construction
    >>> np.dot(rot, [1, 0, 0])
    array([ 0.70710678,  0.        , -0.70710678])
    """

    ux = rot_vector[0]
    uy = rot_vector[1]
    uz = rot_vector[2]
    cost = np.cos(theta)
    sint = np.sin(theta)

    R = np.array(
        [
            [
                cost + ux**2 * (1 - cost),
                ux * uy * (1 - cost) - uz * sint,
                ux * uz * (1 - cost) + uy * sint,
            ],
            [
                uy * ux * (1 - cost) + uz * sint,
                cost + uy**2 * (1 - cost),
                uy * uz * (1 - cost) - ux * sint,
            ],
            [
                uz * ux * (1 - cost) - uy * sint,
                uz * uy * (1 - cost) + ux * sint,
                cost + uz**2 * (1 - cost),
            ],
        ]
    )

    return R


def quaternion_mult(q1, q2):
    """

    Multiply two quaternions. The inputs are 4-component numpy arrays
    in the order [w, x, y, z].

    """
    w = q1[0] * q2[0] - q1[1] * q2[1] - q1[2] * q2[2] - q1[3] * q2[3]
    x = q1[0] * q2[1] + q1[1] * q2[0] + q1[2] * q2[3] - q1[3] * q2[2]
    y = q1[0] * q2[2] + q1[2] * q2[0] + q1[3] * q2[1] - q1[1] * q2[3]
    z = q1[0] * q2[3] + q1[3] * q2[0] + q1[1] * q2[2] - q1[2] * q2[1]
    return np.array([w, x, y, z])


def quaternion_to_rotation_matrix(quaternion):
    """

    This converts a quaternion representation of on orientation to
    a rotation matrix. The input is a 4-component numpy array in
    the order [w, x, y, z], and the output is a 3x3 matrix stored
    as a 2D numpy array.  We follow the approach in
    "3D Math Primer for Graphics and Game Development" by
    Dunn and Parberry.

    """

    w = quaternion[0]
    x = quaternion[1]
    y = quaternion[2]
    z = quaternion[3]

    R = np.empty((3, 3), dtype=np.float64)

    R[0][0] = 1.0 - 2.0 * y**2 - 2.0 * z**2
    R[0][1] = 2.0 * x * y + 2.0 * w * z
    R[0][2] = 2.0 * x * z - 2.0 * w * y

    R[1][0] = 2.0 * x * y - 2.0 * w * z
    R[1][1] = 1.0 - 2.0 * x**2 - 2.0 * z**2
    R[1][2] = 2.0 * y * z + 2.0 * w * x

    R[2][0] = 2.0 * x * z + 2.0 * w * y
    R[2][1] = 2.0 * y * z - 2.0 * w * x
    R[2][2] = 1.0 - 2.0 * x**2 - 2.0 * y**2

    return R


def rotation_matrix_to_quaternion(rot_matrix):
    """

    Convert a rotation matrix-based representation of an
    orientation to a quaternion. The input should be a
    3x3 rotation matrix, while the output will be a
    4-component numpy array. We follow the approach in
    "3D Math Primer for Graphics and Game Development" by
    Dunn and Parberry.

    """
    m11 = rot_matrix[0][0]
    m12 = rot_matrix[0][1]
    m13 = rot_matrix[0][2]
    m21 = rot_matrix[1][0]
    m22 = rot_matrix[1][1]
    m23 = rot_matrix[1][2]
    m31 = rot_matrix[2][0]
    m32 = rot_matrix[2][1]
    m33 = rot_matrix[2][2]

    four_w_squared_minus_1 = m11 + m22 + m33
    four_x_squared_minus_1 = m11 - m22 - m33
    four_y_squared_minus_1 = m22 - m11 - m33
    four_z_squared_minus_1 = m33 - m11 - m22
    max_index = 0
    four_max_squared_minus_1 = four_w_squared_minus_1
    if four_x_squared_minus_1 > four_max_squared_minus_1:
        four_max_squared_minus_1 = four_x_squared_minus_1
        max_index = 1
    if four_y_squared_minus_1 > four_max_squared_minus_1:
        four_max_squared_minus_1 = four_y_squared_minus_1
        max_index = 2
    if four_z_squared_minus_1 > four_max_squared_minus_1:
        four_max_squared_minus_1 = four_z_squared_minus_1
        max_index = 3

    max_val = 0.5 * np.sqrt(four_max_squared_minus_1 + 1.0)
    mult = 0.25 / max_val

    if max_index == 0:
        w = max_val
        x = (m23 - m32) * mult
        y = (m31 - m13) * mult
        z = (m12 - m21) * mult
    elif max_index == 1:
        x = max_val
        w = (m23 - m32) * mult
        y = (m12 + m21) * mult
        z = (m31 + m13) * mult
    elif max_index == 2:
        y = max_val
        w = (m31 - m13) * mult
        x = (m12 + m21) * mult
        z = (m23 + m32) * mult
    elif max_index == 3:
        z = max_val
        w = (m12 - m21) * mult
        x = (m31 + m13) * mult
        y = (m23 + m32) * mult

    return np.array([w, x, y, z])


def get_sph_r(coords):
    # The spherical coordinates radius is simply the magnitude of the
    # coordinate vector.

    return np.sqrt(np.sum(coords**2, axis=0))


def resize_vector(vector, vector_array):
    if len(vector_array.shape) == 4:
        res_vector = np.resize(vector, (3, 1, 1, 1))
    else:
        res_vector = np.resize(vector, (3, 1))
    return res_vector


def normalize_vector(vector):
    # this function normalizes
    # an input vector

    L2 = np.atleast_1d(np.linalg.norm(vector))
    L2[L2 == 0] = 1.0
    vector = vector / L2
    return vector


def get_sph_theta(coords, normal):
    # The angle (theta) with respect to the normal (J), is the arccos
    # of the dot product of the normal with the normalized coordinate
    # vector.

    res_normal = resize_vector(normal, coords)

    # check if the normal vector is normalized
    # since arccos requires the vector to be normalised
    res_normal = normalize_vector(res_normal)

    tile_shape = [1] + list(coords.shape)[1:]

    J = np.tile(res_normal, tile_shape)

    JdotCoords = np.sum(J * coords, axis=0)

    with np.errstate(invalid="ignore"):
        ret = np.arccos(JdotCoords / np.sqrt(np.sum(coords**2, axis=0)))

    ret[np.isnan(ret)] = 0

    return ret


def get_sph_phi(coords, normal):
    # We have freedom with respect to what axis (xprime) to define
    # the disk angle. Here I've chosen to use the axis that is
    # perpendicular to the normal and the y-axis. When normal ==
    # y-hat, then set xprime = z-hat. With this definition, when
    # normal == z-hat (as is typical), then xprime == x-hat.
    #
    # The angle is then given by the arctan of the ratio of the
    # yprime-component and the xprime-component of the coordinate
    # vector.

    normal = normalize_vector(normal)
    (zprime, xprime, yprime) = ortho_find(normal)

    res_xprime = resize_vector(xprime, coords)
    res_yprime = resize_vector(yprime, coords)

    tile_shape = [1] + list(coords.shape)[1:]
    Jx = np.tile(res_xprime, tile_shape)
    Jy = np.tile(res_yprime, tile_shape)

    Px = np.sum(Jx * coords, axis=0)
    Py = np.sum(Jy * coords, axis=0)

    return np.arctan2(Py, Px)


def get_cyl_r(coords, normal):
    # The cross product of the normal (J) with a coordinate vector
    # gives a vector of magnitude equal to the cylindrical radius.

    res_normal = resize_vector(normal, coords)
    res_normal = normalize_vector(res_normal)

    tile_shape = [1] + list(coords.shape)[1:]
    J = np.tile(res_normal, tile_shape)

    JcrossCoords = np.cross(J, coords, axisa=0, axisb=0, axisc=0)
    return np.sqrt(np.sum(JcrossCoords**2, axis=0))


def get_cyl_z(coords, normal):
    # The dot product of the normal (J) with the coordinate vector
    # gives the cylindrical height.

    res_normal = resize_vector(normal, coords)
    res_normal = normalize_vector(res_normal)

    tile_shape = [1] + list(coords.shape)[1:]
    J = np.tile(res_normal, tile_shape)

    return np.sum(J * coords, axis=0)


def get_cyl_theta(coords, normal):
    # This is identical to the spherical phi component

    return get_sph_phi(coords, normal)


def get_cyl_r_component(vectors, theta, normal):
    # The r of a vector is the vector dotted with rhat

    normal = normalize_vector(normal)
    (zprime, xprime, yprime) = ortho_find(normal)

    res_xprime = resize_vector(xprime, vectors)
    res_yprime = resize_vector(yprime, vectors)

    tile_shape = [1] + list(vectors.shape)[1:]
    Jx = np.tile(res_xprime, tile_shape)
    Jy = np.tile(res_yprime, tile_shape)

    rhat = Jx * np.cos(theta) + Jy * np.sin(theta)

    return np.sum(vectors * rhat, axis=0)


def get_cyl_theta_component(vectors, theta, normal):
    # The theta component of a vector is the vector dotted with thetahat
    normal = normalize_vector(normal)
    (zprime, xprime, yprime) = ortho_find(normal)

    res_xprime = resize_vector(xprime, vectors)
    res_yprime = resize_vector(yprime, vectors)

    tile_shape = [1] + list(vectors.shape)[1:]
    Jx = np.tile(res_xprime, tile_shape)
    Jy = np.tile(res_yprime, tile_shape)

    thetahat = -Jx * np.sin(theta) + Jy * np.cos(theta)

    return np.sum(vectors * thetahat, axis=0)


def get_cyl_z_component(vectors, normal):
    # The z component of a vector is the vector dotted with zhat
    normal = normalize_vector(normal)
    (zprime, xprime, yprime) = ortho_find(normal)

    res_zprime = resize_vector(zprime, vectors)

    tile_shape = [1] + list(vectors.shape)[1:]
    zhat = np.tile(res_zprime, tile_shape)

    return np.sum(vectors * zhat, axis=0)


def get_sph_r_component(vectors, theta, phi, normal):
    # The r component of a vector is the vector dotted with rhat
    normal = normalize_vector(normal)
    (zprime, xprime, yprime) = ortho_find(normal)

    res_xprime = resize_vector(xprime, vectors)
    res_yprime = resize_vector(yprime, vectors)
    res_zprime = resize_vector(zprime, vectors)

    tile_shape = [1] + list(vectors.shape)[1:]

    Jx, Jy, Jz = (
        YTArray(np.tile(rprime, tile_shape), "")
        for rprime in (res_xprime, res_yprime, res_zprime)
    )

    rhat = (
        Jx * np.sin(theta) * np.cos(phi)
        + Jy * np.sin(theta) * np.sin(phi)
        + Jz * np.cos(theta)
    )

    return np.sum(vectors * rhat, axis=0)


def get_sph_phi_component(vectors, phi, normal):
    # The phi component of a vector is the vector dotted with phihat
    normal = normalize_vector(normal)
    (zprime, xprime, yprime) = ortho_find(normal)

    res_xprime = resize_vector(xprime, vectors)
    res_yprime = resize_vector(yprime, vectors)

    tile_shape = [1] + list(vectors.shape)[1:]
    Jx = YTArray(np.tile(res_xprime, tile_shape), "")
    Jy = YTArray(np.tile(res_yprime, tile_shape), "")

    phihat = -Jx * np.sin(phi) + Jy * np.cos(phi)

    return np.sum(vectors * phihat, axis=0)


def get_sph_theta_component(vectors, theta, phi, normal):
    # The theta component of a vector is the vector dotted with thetahat
    normal = normalize_vector(normal)
    (zprime, xprime, yprime) = ortho_find(normal)

    res_xprime = resize_vector(xprime, vectors)
    res_yprime = resize_vector(yprime, vectors)
    res_zprime = resize_vector(zprime, vectors)

    tile_shape = [1] + list(vectors.shape)[1:]
    Jx, Jy, Jz = (
        YTArray(np.tile(rprime, tile_shape), "")
        for rprime in (res_xprime, res_yprime, res_zprime)
    )

    thetahat = (
        Jx * np.cos(theta) * np.cos(phi)
        + Jy * np.cos(theta) * np.sin(phi)
        - Jz * np.sin(theta)
    )

    return np.sum(vectors * thetahat, axis=0)


def compute_stddev_image(buff2, buff):
    """
    This function computes the standard deviation of a weighted projection.
    It defines the standard deviation as sigma^2 = <v^2>-<v>^2, where the
    brackets indicate averages (with the weight) and v is the field in
    question.

    There is the possibility that if the projection at a particular location
    is of a constant or a single cell/particle, then <v^2> = <v>^2 and instead
    of getting zero one gets roundoff error that results in sigma^2 < 0,
    which is unphysical.

    We handle this here by performing the subtraction and checking that any
    and all negative values of sigma^2 can be attributed to roundoff and
    thus be safely set to zero. We error out if this is not the case. There
    are ways of computing the standard deviation that are designed to avoid
    this catastrophic cancellation, but this would require rather substantial
    and invasive changes to the projection machinery so for the time being
    it is avoided.

    Parameters
    ----------
    buff2 : ImageArray
        The image of the weighted averge of the field squared
    buff : ImageArray
        The image of the weighted averge of the field
    """
    buff1 = buff * buff
    y = buff2 - buff1
    close_negs = np.isclose(np.asarray(buff2), np.asarray(buff1))[y < 0.0]
    if close_negs.all():
        y[y < 0.0] = 0.0
    else:
        raise ValueError(
            "Something went wrong, there are significant negative "
            "variances in the standard deviation image!"
        )
    return np.sqrt(y)



import os
import shutil
import tempfile
import unittest

import pytest

import yt
from yt.testing import assert_allclose_units, fake_random_ds
from yt.visualization.api import PhasePlot


class TestPhasePlotAPI:
    @classmethod
    def setup_class(cls):
        cls.ds = fake_random_ds(
            16, fields=("density", "temperature"), units=("g/cm**3", "K")
        )

    def get_plot(self):
        return PhasePlot(
            self.ds, ("gas", "density"), ("gas", "temperature"), ("gas", "mass")
        )

    @pytest.mark.parametrize("kwargs", [{}, {"color": "b"}])
    @pytest.mark.mpl_image_compare
    def test_phaseplot_annotate_text(self, kwargs):
        p = self.get_plot()
        p.annotate_text(1e-4, 1e-2, "Test text annotation", **kwargs)
        p.render()
        return p.plots["gas", "mass"].figure

    @pytest.mark.mpl_image_compare
    def test_phaseplot_set_title(self):
        p = self.get_plot()
        p.set_title(("gas", "mass"), "Test Title")
        p.render()
        return p.plots["gas", "mass"].figure

    @pytest.mark.mpl_image_compare
    def test_phaseplot_set_log(self):
        p = self.get_plot()
        p.set_log(("gas", "mass"), False)
        p.render()
        return p.plots["gas", "mass"].figure

    @pytest.mark.mpl_image_compare
    def test_phaseplot_set_unit(self):
        p = self.get_plot()
        p.set_unit(("gas", "mass"), "Msun")
        p.render()
        return p.plots["gas", "mass"].figure

    @pytest.mark.mpl_image_compare
    def test_phaseplot_set_xlim(self):
        p = self.get_plot()
        p.set_xlim(1e-3, 1e0)
        p.render()
        return p.plots["gas", "mass"].figure

    @pytest.mark.mpl_image_compare
    def test_phaseplot_set_ylim(self):
        p = self.get_plot()
        p.set_ylim(1e-2, 1e0)
        p.render()
        return p.plots["gas", "mass"].figure


def test_set_units():
    fields = ("density", "temperature")
    units = (
        "g/cm**3",
        "K",
    )
    ds = fake_random_ds(16, fields=fields, units=units)
    sp = ds.sphere("max", (1.0, "Mpc"))
    p1 = yt.ProfilePlot(sp, ("index", "radius"), ("gas", "density"))
    p2 = yt.PhasePlot(sp, ("gas", "density"), ("gas", "temperature"), ("gas", "mass"))
    # make sure we can set the units using the tuple without erroring out
    p1.set_unit(("gas", "density"), "Msun/kpc**3")
    p2.set_unit(("gas", "temperature"), "R")


def test_set_labels():
    ds = fake_random_ds(16)
    ad = ds.all_data()
    plot = yt.ProfilePlot(
        ad,
        ("index", "radius"),
        [("gas", "velocity_x"), ("gas", "density")],
        weight_field=None,
    )
    # make sure we can set the labels without erroring out
    plot.set_ylabel("all", "test ylabel")
    plot.set_xlabel("test xlabel")


def test_create_from_dataset():
    ds = fake_random_ds(16)
    plot1 = yt.ProfilePlot(
        ds,
        ("index", "radius"),
        [("gas", "velocity_x"), ("gas", "density")],
        weight_field=None,
    )
    plot2 = yt.ProfilePlot(
        ds.all_data(),
        ("index", "radius"),
        [("gas", "velocity_x"), ("gas", "density")],
        weight_field=None,
    )
    assert_allclose_units(
        plot1.profiles[0][("gas", "density")], plot2.profiles[0][("gas", "density")]
    )
    assert_allclose_units(
        plot1.profiles[0]["velocity_x"], plot2.profiles[0]["velocity_x"]
    )

    plot1 = yt.PhasePlot(ds, ("gas", "density"), ("gas", "velocity_x"), ("gas", "mass"))
    plot2 = yt.PhasePlot(
        ds.all_data(), ("gas", "density"), ("gas", "velocity_x"), ("gas", "mass")
    )
    assert_allclose_units(plot1.profile["mass"], plot2.profile["mass"])


class TestAnnotations(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tmpdir = tempfile.mkdtemp()
        cls.curdir = os.getcwd()
        os.chdir(cls.tmpdir)

        ds = fake_random_ds(16)
        ad = ds.all_data()
        cls.fields = [
            ("gas", "velocity_x"),
            ("gas", "velocity_y"),
            ("gas", "velocity_z"),
        ]
        cls.plot = yt.ProfilePlot(
            ad, ("index", "radius"), cls.fields, weight_field=None
        )

    @classmethod
    def tearDownClass(cls):
        os.chdir(cls.curdir)
        shutil.rmtree(cls.tmpdir)

    def test_annotations(self):
        # make sure we can annotate without erroring out
        # annotate the plot with only velocity_x
        self.plot.annotate_title("velocity_x plot", self.fields[0])
        self.plot.annotate_text(1e-1, 1e1, "Annotated velocity_x")

        # annotate the plots with velocity_y and velocity_z with
        # the same annotations
        self.plot.annotate_title("Velocity Plots (Y or Z)", self.fields[1:])
        self.plot.annotate_text(1e-1, 1e1, "Annotated vel_y, vel_z", self.fields[1:])
        self.plot.save()

    def test_annotations_wrong_fields(self):
        from yt.utilities.exceptions import YTFieldNotFound

        with self.assertRaises(YTFieldNotFound):
            self.plot.annotate_title("velocity_x plot", "wrong_field_name")

        with self.assertRaises(YTFieldNotFound):
            self.plot.annotate_text(1e-1, 1e1, "Annotated text", "wrong_field_name")


def test_phaseplot_set_log():
    fields = ("density", "temperature")
    units = (
        "g/cm**3",
        "K",
    )
    ds = fake_random_ds(16, fields=fields, units=units)
    sp = ds.sphere("max", (1.0, "Mpc"))
    p1 = yt.ProfilePlot(sp, ("index", "radius"), ("gas", "density"))
    p2 = yt.PhasePlot(sp, ("gas", "density"), ("gas", "temperature"), ("gas", "mass"))
    # make sure we can set the log-scaling using the tuple without erroring out
    p1.set_log(("gas", "density"), False)
    p2.set_log(("gas", "temperature"), False)
    assert not p1.y_log["gas", "density"]
    assert not p2.y_log

    # make sure we can set the log-scaling using a string without erroring out
    p1.set_log(("gas", "density"), True)
    p2.set_log(("gas", "temperature"), True)
    assert p1.y_log["gas", "density"]
    assert p2.y_log

    # make sure we can set the log-scaling using a field object
    p1.set_log(ds.fields.gas.density, False)
    p2.set_log(ds.fields.gas.temperature, False)
    assert not p1.y_log["gas", "density"]
    assert not p2.y_log


def test_phaseplot_showhide_colorbar_axes():
    fields = ("density", "temperature")
    units = (
        "g/cm**3",
        "K",
    )
    ds = fake_random_ds(16, fields=fields, units=units)
    ad = ds.all_data()
    plot = yt.PhasePlot(ad, ("gas", "density"), ("gas", "temperature"), ("gas", "mass"))

    # make sure we can hide colorbar
    plot.hide_colorbar()
    with tempfile.NamedTemporaryFile(suffix="png") as f1:
        plot.save(f1.name)

    # make sure we can show colorbar
    plot.show_colorbar()
    with tempfile.NamedTemporaryFile(suffix="png") as f2:
        plot.save(f2.name)

    # make sure we can hide axes
    plot.hide_axes()
    with tempfile.NamedTemporaryFile(suffix="png") as f3:
        plot.save(f3.name)

    # make sure we can show axes
    plot.show_axes()
    with tempfile.NamedTemporaryFile(suffix="png") as f4:
        plot.save(f4.name)



from numpy.testing import assert_equal

from yt.testing import requires_file, requires_module, units_override_check
from yt.utilities.answer_testing.framework import (
    data_dir_load,
    requires_ds,
    small_patch_amr,
)

from ..data_structures import (
    EventsFITSDataset,
    FITSDataset,
    SkyDataFITSDataset,
    SpectralCubeFITSDataset,
)

_fields_grs = (("fits", "temperature"),)

grs = "radio_fits/grs-50-cube.fits"


@requires_ds(grs)
def test_grs():
    ds = data_dir_load(grs, cls=SpectralCubeFITSDataset, kwargs={"nan_mask": 0.0})
    assert_equal(str(ds), "grs-50-cube.fits")
    for test in small_patch_amr(ds, _fields_grs, input_center="c", input_weight="ones"):
        test_grs.__name__ = test.description
        yield test


_fields_vels = (("fits", "velocity_x"), ("fits", "velocity_y"), ("fits", "velocity_z"))

vf = "UnigridData/velocity_field_20.fits"


@requires_module("astropy")
@requires_ds(vf)
def test_velocity_field():
    ds = data_dir_load(vf, cls=FITSDataset)
    assert_equal(str(ds), "velocity_field_20.fits")
    for test in small_patch_amr(
        ds, _fields_vels, input_center="c", input_weight="ones"
    ):
        test_velocity_field.__name__ = test.description
        yield test


acis = "xray_fits/acisf05356N003_evt2.fits.gz"

_fields_acis = (("gas", "counts_0.1-2.0"), ("gas", "counts_2.0-5.0"))


@requires_ds(acis)
def test_acis():
    from yt.frontends.fits.misc import setup_counts_fields

    ds = data_dir_load(acis, cls=EventsFITSDataset)
    ebounds = [(0.1, 2.0), (2.0, 5.0)]
    setup_counts_fields(ds, ebounds)
    assert_equal(str(ds), "acisf05356N003_evt2.fits.gz")
    for test in small_patch_amr(
        ds, _fields_acis, input_center="c", input_weight="ones"
    ):
        test_acis.__name__ = test.description
        yield test


A2052 = "xray_fits/A2052_merged_0.3-2_match-core_tmap_bgecorr.fits"

_fields_A2052 = (("fits", "flux"),)


@requires_ds(A2052)
def test_A2052():
    ds = data_dir_load(A2052, cls=SkyDataFITSDataset)
    assert_equal(str(ds), "A2052_merged_0.3-2_match-core_tmap_bgecorr.fits")
    for test in small_patch_amr(
        ds, _fields_A2052, input_center="c", input_weight="ones"
    ):
        test_A2052.__name__ = test.description
        yield test


@requires_file(vf)
def test_units_override():
    units_override_check(vf)


@requires_file(vf)
def test_FITSDataset():
    assert isinstance(data_dir_load(vf), FITSDataset)


@requires_file(grs)
def test_SpectralCubeFITSDataset():
    assert isinstance(data_dir_load(grs), SpectralCubeFITSDataset)


@requires_file(acis)
def test_EventsFITSDataset():
    assert isinstance(data_dir_load(acis), EventsFITSDataset)


@requires_file(A2052)
def test_SkyDataFITSDataset():
    assert isinstance(data_dir_load(A2052), SkyDataFITSDataset)



import numpy as np

from yt.loaders import load_uniform_grid
from numpy.testing import assert_allclose


def test_magnetic_code_units():

    sqrt4pi = np.sqrt(4.0 * np.pi)
    ddims = (16,) * 3
    data = {"density": (np.random.uniform(size=ddims), "g/cm**3")}

    ds1 = load_uniform_grid(
        data, ddims, magnetic_unit=(sqrt4pi, "gauss"), unit_system="cgs"
    )

    assert_allclose(ds1.magnetic_unit.value, sqrt4pi)
    assert str(ds1.magnetic_unit.units) == "G"

    mucu = ds1.magnetic_unit.to("code_magnetic")
    assert_allclose(mucu.value, 1.0)
    assert str(mucu.units) == "code_magnetic"

    ds2 = load_uniform_grid(data, ddims, magnetic_unit=(1.0, "T"), unit_system="cgs")
    assert_allclose(ds2.magnetic_unit.value, 10000.0)
    assert str(ds2.magnetic_unit.units) == "G"

    mucu = ds2.magnetic_unit.to("code_magnetic")
    assert_allclose(mucu.value, 1.0)
    assert str(mucu.units) == "code_magnetic"

    ds3 = load_uniform_grid(data, ddims, magnetic_unit=(1.0, "T"), unit_system="mks")

    assert_allclose(ds3.magnetic_unit.value, 1.0)
    assert str(ds3.magnetic_unit.units) == "T"

    mucu = ds3.magnetic_unit.to("code_magnetic")
    assert_allclose(mucu.value, 1.0)
    assert str(mucu.units) == "code_magnetic"

    ds4 = load_uniform_grid(
        data, ddims, magnetic_unit=(1.0, "gauss"), unit_system="mks"
    )

    assert_allclose(ds4.magnetic_unit.value, 1.0e-4)
    assert str(ds4.magnetic_unit.units) == "T"

    mucu = ds4.magnetic_unit.to("code_magnetic")
    assert_allclose(mucu.value, 1.0)
    assert str(mucu.units) == "code_magnetic"

    ds5 = load_uniform_grid(
        data, ddims, magnetic_unit=(1.0, "uG"), unit_system="mks"
    )

    assert_allclose(ds5.magnetic_unit.value, 1.0e-10)
    assert str(ds5.magnetic_unit.units) == "T"

    mucu = ds5.magnetic_unit.to("code_magnetic")
    assert_allclose(mucu.value, 1.0)
    assert str(mucu.units) == "code_magnetic"

    ds6 = load_uniform_grid(
        data, ddims, magnetic_unit=(1.0, "nT"), unit_system="cgs"
    )

    assert_allclose(ds6.magnetic_unit.value, 1.0e-5)
    assert str(ds6.magnetic_unit.units) == "G"

    mucu = ds6.magnetic_unit.to("code_magnetic")
    assert_allclose(mucu.value, 1.0)
    assert str(mucu.units) == "code_magnetic"



import os.path

from numpy.testing import assert_equal

from yt.frontends.rockstar.api import RockstarDataset
from yt.testing import ParticleSelectionComparison, requires_file
from yt.utilities.answer_testing.framework import (
    FieldValuesTest,
    data_dir_load,
    requires_ds,
)

_fields = (
    ("all", "particle_position_x"),
    ("all", "particle_position_y"),
    ("all", "particle_position_z"),
    ("all", "particle_mass"),
)

r1 = "rockstar_halos/halos_0.0.bin"


@requires_ds(r1)
def test_fields_r1():
    ds = data_dir_load(r1)
    assert_equal(str(ds), os.path.basename(r1))
    for field in _fields:
        yield FieldValuesTest(r1, field, particle_type=True)


@requires_file(r1)
def test_RockstarDataset():
    assert isinstance(data_dir_load(r1), RockstarDataset)


@requires_file(r1)
def test_particle_selection():
    ds = data_dir_load(r1)
    psc = ParticleSelectionComparison(ds)
    psc.run_defaults()



import numpy as np
from numpy.testing import assert_almost_equal, assert_equal

from yt import SlicePlot, add_particle_filter
from yt.loaders import load
from yt.testing import fake_sph_grid_ds, fake_sph_orientation_ds, requires_file


def test_point():
    ds = fake_sph_orientation_ds()
    field_data = ds.stream_handler.fields["stream_file"]
    ppos = [field_data["io", f"particle_position_{d}"] for d in "xyz"]
    ppos = np.array(ppos).T
    for pos in ppos:
        for i in range(-1, 2):
            offset = 0.1 * np.array([i, 0, 0])
            pt = ds.point(pos + offset)
            assert_equal(pt["gas", "density"].shape[0], 1)
        for j in range(-1, 2):
            offset = 0.1 * np.array([0, j, 0])
            pt = ds.point(pos + offset)
            assert_equal(pt["gas", "density"].shape[0], 1)
        for k in range(-1, 2):
            offset = 0.1 * np.array([0, 0, k])
            pt = ds.point(pos + offset)
            assert_equal(pt["gas", "density"].shape[0], 1)


# The number of particles along each slice axis at that coordinate
SLICE_ANSWERS = {
    ("x", 0): 6,
    ("x", 0.5): 0,
    ("x", 1): 1,
    ("y", 0): 5,
    ("y", 1): 1,
    ("y", 2): 1,
    ("z", 0): 4,
    ("z", 1): 1,
    ("z", 2): 1,
    ("z", 3): 1,
}


def test_slice():
    ds = fake_sph_orientation_ds()
    for (ax, coord), answer in SLICE_ANSWERS.items():
        # test that we can still select particles even if we offset the slice
        # within each particle's smoothing volume
        for i in range(-1, 2):
            sl = ds.slice(ax, coord + i * 0.1)
            assert_equal(sl["gas", "density"].shape[0], answer)


REGION_ANSWERS = {
    ((-4, -4, -4), (4, 4, 4)): 7,
    ((0, 0, 0), (4, 4, 4)): 7,
    ((1, 0, 0), (4, 4, 4)): 1,
    ((0, 1, 0), (4, 4, 4)): 2,
    ((0, 0, 1), (4, 4, 4)): 3,
    ((0, 0, 0), (4, 4, 2)): 6,
    ((0, 0, 0), (4, 4, 1)): 5,
    ((0, 0, 0), (4, 1, 4)): 6,
    ((0, 0, 0), (1, 1, 4)): 6,
}


def test_region():
    ds = fake_sph_orientation_ds()
    for (left_edge, right_edge), answer in REGION_ANSWERS.items():
        # test that regions enclosing a particle's smoothing region
        # correctly select SPH particles
        for i in range(-1, 2):
            for j in range(-1, 2):
                le = np.array([le + i * 0.1 for le in left_edge])
                re = np.array([re + j * 0.1 for re in right_edge])

                # check if we went off the edge of the domain
                whl = le < ds.domain_left_edge
                le[whl] = ds.domain_left_edge[whl]
                whr = re > ds.domain_right_edge
                re[whr] = ds.domain_right_edge[whr]

                reg = ds.box(le, re)
                assert_equal(reg["gas", "density"].shape[0], answer)


def test_periodic_region():
    ds = fake_sph_grid_ds(10.0)
    coords = [0.7, 1.4, 2.8]

    for x in coords:
        for y in coords:
            for z in coords:
                center = np.array([x, y, z])
                for n, w in zip((8, 27), (1.0, 2.0)):
                    le = center - 0.5 * w
                    re = center + 0.5 * w
                    box = ds.box(le, re)
                    assert box["io", "particle_ones"].size == n


SPHERE_ANSWERS = {
    ((0, 0, 0), 4): 7,
    ((0, 0, 0), 3): 7,
    ((0, 0, 0), 2): 6,
    ((0, 0, 0), 1): 4,
    ((0, 0, 0), 0.5): 1,
    ((1, 0, 0), 0.5): 1,
    ((1, 0, 0), 1.0): 2,
    ((0, 1, 0), 1.0): 3,
    ((0, 0, 1), 1.0): 3,
}


def test_sphere():
    ds = fake_sph_orientation_ds()
    for (center, radius), answer in SPHERE_ANSWERS.items():
        # test that spheres enclosing a particle's smoothing region
        # correctly select SPH particles
        for i in range(-1, 2):
            for j in range(-1, 2):
                cent = np.array([c + i * 0.1 for c in center])
                rad = radius + 0.1 * j
                sph = ds.sphere(cent, rad)
                assert_equal(sph["gas", "density"].shape[0], answer)


DISK_ANSWERS = {
    ((0, 0, 0), (0, 0, 1), 4, 3): 7,
    ((0, 0, 0), (0, 0, 1), 4, 2): 6,
    ((0, 0, 0), (0, 0, 1), 4, 1): 5,
    ((0, 0, 0), (0, 0, 1), 4, 0.5): 4,
    ((0, 0, 0), (0, 1, 0), 4, 3): 7,
    ((0, 0, 0), (0, 1, 0), 4, 2): 7,
    ((0, 0, 0), (0, 1, 0), 4, 1): 6,
    ((0, 0, 0), (0, 1, 0), 4, 0.5): 5,
    ((0, 0, 0), (1, 0, 0), 4, 3): 7,
    ((0, 0, 0), (1, 0, 0), 4, 2): 7,
    ((0, 0, 0), (1, 0, 0), 4, 1): 7,
    ((0, 0, 0), (1, 0, 0), 4, 0.5): 6,
    ((0, 0, 0), (1, 1, 1), 1, 1): 4,
    ((-0.5, -0.5, -0.5), (1, 1, 1), 4, 4): 7,
}


def test_disk():
    ds = fake_sph_orientation_ds()
    for (center, normal, radius, height), answer in DISK_ANSWERS.items():
        # test that disks enclosing a particle's smoothing region
        # correctly select SPH particles
        for i in range(-1, 2):
            cent = np.array([c + i * 0.1 for c in center])
            disk = ds.disk(cent, normal, radius, height)
            assert_equal(disk["gas", "density"].shape[0], answer)


RAY_ANSWERS = {
    ((0, 0, 0), (3, 0, 0)): 2,
    ((0, 0, 0), (0, 3, 0)): 3,
    ((0, 0, 0), (0, 0, 3)): 4,
    ((0, 1, 0), (0, 2, 0)): 2,
    ((1, 0, 0), (0, 2, 0)): 2,
    ((0.5, 0.5, 0.5), (0.5, 0.5, 3.5)): 0,
}


def test_ray():
    ds = fake_sph_orientation_ds()
    for (start_point, end_point), answer in RAY_ANSWERS.items():
        for i in range(-1, 2):
            start = np.array([s + i * 0.1 for s in start_point])
            end = np.array([e + i * 0.1 for e in end_point])
            ray = ds.ray(start, end)
            assert_equal(ray["gas", "density"].shape[0], answer)


CUTTING_ANSWERS = {
    ((1, 0, 0), (0, 0, 0)): 6,
    ((0, 1, 0), (0, 0, 0)): 5,
    ((0, 0, 1), (0, 0, 0)): 4,
    ((1, 1, 1), (1.0 / 3, 1.0 / 3, 1.0 / 3)): 3,
    ((1, 1, 1), (2.0 / 3, 2.0 / 3, 2.0 / 3)): 2,
    ((1, 1, 1), (1, 1, 1)): 1,
}


def test_cutting():
    ds = fake_sph_orientation_ds()
    for (normal, center), answer in CUTTING_ANSWERS.items():
        for i in range(-1, 2):
            cen = [c + 0.1 * i for c in center]
            cut = ds.cutting(normal, cen)
            assert_equal(cut["gas", "density"].shape[0], answer)


def test_chained_selection():
    ds = fake_sph_orientation_ds()

    for (center, radius), answer in SPHERE_ANSWERS.items():
        sph = ds.sphere(center, radius)
        region = ds.box(ds.domain_left_edge, ds.domain_right_edge, data_source=sph)
        assert_equal(region["gas", "density"].shape[0], answer)


def test_boolean_selection():
    ds = fake_sph_orientation_ds()

    sph = ds.sphere([0, 0, 0], 0.5)

    sph2 = ds.sphere([1, 0, 0], 0.5)

    reg = ds.all_data()

    neg = reg - sph

    assert_equal(neg["gas", "density"].shape[0], 6)

    plus = sph + sph2

    assert_equal(plus["gas", "density"].shape[0], 2)

    intersect = sph & sph2

    assert_equal(intersect["gas", "density"].shape[0], 0)

    intersect = reg & sph2

    assert_equal(intersect["gas", "density"].shape[0], 1)

    exclusive = sph ^ sph2

    assert_equal(exclusive["gas", "density"].shape[0], 2)

    exclusive = sph ^ reg

    assert_equal(exclusive["gas", "density"].shape[0], 6)

    intersect = ds.intersection([sph, sph2])

    assert_equal(intersect["gas", "density"].shape[0], 0)

    intersect = ds.intersection([reg, sph2])

    assert_equal(intersect["gas", "density"].shape[0], 1)

    union = ds.union([sph, sph2])

    assert_equal(union["gas", "density"].shape[0], 2)

    union = ds.union([sph, reg])

    assert_equal(union["gas", "density"].shape[0], 7)


def test_arbitrary_grid():
    ds = fake_sph_grid_ds()

    # this loads up some sph data in a test grid
    agrid = ds.arbitrary_grid([0, 0, 0], [3, 3, 3], dims=[3, 3, 3])

    # the field should be equal to the density of a particle in every voxel
    # which is 1.
    dens = agrid["gas", "density"]
    answers = np.ones(shape=(3, 3, 3))

    assert_equal(dens, answers)


def test_compare_arbitrary_grid_slice():
    ds = fake_sph_orientation_ds()
    c = np.array([0.0, 0.0, 0.0])
    width = 1.5
    buff_size = 51
    field = ("gas", "density")

    # buffer from arbitrary grid
    ag = ds.arbitrary_grid(c - width / 2, c + width / 2, [buff_size] * 3)
    buff_ag = ag[field][:, :, int(np.floor(buff_size / 2))].d.T

    # buffer from slice
    p = SlicePlot(ds, "z", field, center=c, width=width)
    p.set_buff_size(51)
    buff_slc = p.frb.data[field].d

    assert_equal(buff_slc, buff_ag)


def test_gather_slice():
    ds = fake_sph_grid_ds()
    ds.num_neighbors = 5
    field = ("gas", "density")

    c = np.array([1.5, 1.5, 0.5])
    width = 3.0

    p = SlicePlot(ds, "z", field, center=c, width=width)
    p.set_buff_size(3)
    buff_scatter = p.frb.data[field].d

    ds.sph_smoothing_style = "gather"

    p = SlicePlot(ds, "z", field, center=c, width=width)
    p.set_buff_size(3)
    buff_gather = p.frb.data[field].d

    assert_equal(buff_scatter, buff_gather)


def test_gather_grid():
    ds = fake_sph_grid_ds()
    ds.num_neighbors = 5
    field = ("gas", "density")

    ag = ds.arbitrary_grid([0, 0, 0], [3, 3, 3], dims=[3, 3, 3])
    scatter = ag[field]

    ds.sph_smoothing_style = "gather"
    ag = ds.arbitrary_grid([0, 0, 0], [3, 3, 3], dims=[3, 3, 3])
    gather = ag[field]

    assert_equal(gather, scatter)


def test_covering_grid_scatter():
    ds = fake_sph_grid_ds()
    field = ("gas", "density")
    buff_size = 8

    ag = ds.arbitrary_grid(0, 3, [buff_size] * 3)
    ag_dens = ag[field].to("g*cm**-3").d

    cg = ds.covering_grid(3, 0, 8)
    cg_dens = cg[field].to("g*cm**-3").d

    assert_equal(ag_dens, cg_dens)


def test_covering_grid_gather():
    ds = fake_sph_grid_ds()
    ds.sph_smoothing_style = "gather"
    ds.num_neighbors = 5
    field = ("gas", "density")
    buff_size = 8

    ag = ds.arbitrary_grid(0, 3, [buff_size] * 3)
    ag_dens = ag[field].to("g*cm**-3").d

    cg = ds.covering_grid(3, 0, 8)
    cg_dens = cg[field].to("g*cm**-3").d

    assert_equal(ag_dens, cg_dens)


@requires_file("TNGHalo/halo_59.hdf5")
def test_covering_grid_derived_fields():
    def hot_gas(pfilter, data):
        return data[pfilter.filtered_type, "temperature"] > 1.0e6

    add_particle_filter(
        "hot_gas",
        function=hot_gas,
        filtered_type="gas",
        requires=["temperature"],
    )
    bbox = [[40669.34, 56669.34], [45984.04, 61984.04], [54114.9, 70114.9]]
    ds = load("TNGHalo/halo_59.hdf5", bounding_box=bbox)
    ds.add_particle_filter("hot_gas")
    w = ds.quan(0.2, "Mpc")
    le = ds.domain_center - 0.5 * w
    re = ds.domain_center + 0.5 * w
    g = ds.r[le[0] : re[0] : 128j, le[1] : re[1] : 128j, le[2] : re[2] : 128j]
    T1 = g["gas", "temperature"].to("keV", "thermal")
    T2 = g["gas", "kT"]
    assert_almost_equal(T1, T2)
    T3 = g["hot_gas", "temperature"].to("keV", "thermal")
    T4 = g["hot_gas", "kT"]
    assert_almost_equal(T3, T4)




# distutils: libraries = STD_LIBS
"""
An allocation container and memory pool



"""


cimport numpy as np

import numpy as np


cdef class ObjectPool:
    def __cinit__(self):
        """This class is *not* meant to be initialized directly, but instead
        through subclasses.  Those subclasses need to implement at a minimum
        the setting of itemsize, but can optionally also implement setup_objs
        and teardown_objs to either set default values or initialize additional
        pointers and values, and then free them."""
        self.containers = NULL
        self.n_con = 0
        self.itemsize = -1 # Never use the base class

    def __dealloc__(self):
        cdef int i
        cdef AllocationContainer *obj
        for i in range(self.n_con):
            obj = &self.containers[i]
            self.teardown_objs(obj.my_objs, obj.n, obj.offset, obj.con_id)
        if self.containers != NULL:
            free(self.containers)

    def __getitem__(self, int i):
        """This returns an array view (if possible and implemented in a
        subclass) on the pool of objects specified by i."""
        return self._con_to_array(i)

    def __len__(self):
        return self.n_con

    def append(self, int n_objs, np.int64_t con_id = -1):
        """This allocates a new batch of n_objs objects, with the container id
        specified as con_id.  Return value is a view on them."""
        self.allocate_objs(n_objs, con_id)
        return self[self.n_con - 1]

    cdef void allocate_objs(self, int n_objs, np.int64_t con_id = -1) except *:
        cdef AllocationContainer *n_cont
        cdef AllocationContainer *prev
        self.containers = <AllocationContainer*> realloc(
              self.containers,
              sizeof(AllocationContainer) * (self.n_con + 1))
        n_cont = &self.containers[self.n_con]
        if self.n_con == 0:
            n_cont.offset = 0
        else:
            prev = &self.containers[self.n_con - 1]
            n_cont.offset = prev.offset + prev.n
        self.n_con += 1
        n_cont.my_objs = malloc(self.itemsize * n_objs)
        if n_cont.my_objs == NULL:
            raise MemoryError
        n_cont.n = n_objs
        n_cont.n_assigned = 0
        n_cont.con_id = con_id
        self.setup_objs(n_cont.my_objs, n_objs, n_cont.offset, n_cont.con_id)

    cdef void setup_objs(self, void *obj, np.uint64_t count,
                         np.uint64_t offset, np.int64_t con_id):
        """This can be overridden in the subclass, where it can initialize any
        of the allocated objects."""
        pass

    cdef void teardown_objs(self, void *obj, np.uint64_t n, np.uint64_t offset,
                           np.int64_t con_id):
        # We assume that these are all allocated and have no sub-allocations
        """If overridden, additional behavior can be provided during teardown
        of allocations.  For instance, if you allocate some memory on each of
        the allocated objects."""
        if obj != NULL:
            free(obj)

    def to_arrays(self):
        rv = []
        cdef int i
        for i in range(self.n_con):
            rv.append(self._con_to_array(i))
        return rv

    def _con_to_array(self, int i):
        """This has to be implemented in a subclass, and should return an
        appropriate np.asarray() of a memoryview of self.my_objs."""
        raise NotImplementedError

cdef class BitmaskPool(ObjectPool):
    def __cinit__(self):
        """This is an example implementation of object pools for bitmasks
        (uint8) arrays.  It lets you reasonably quickly allocate a set of
        uint8 arrays which can be accessed and modified, and then virtually
        append to that."""
        # Base class will ALSO be called
        self.itemsize = sizeof(np.uint8_t)

    cdef void setup_objs(self, void *obj, np.uint64_t n, np.uint64_t offset,
                         np.int64_t con_id):
        cdef np.uint64_t i
        cdef np.uint8_t *mask = <np.uint8_t *> obj
        for i in range(n):
            mask[i] = 0

    def _con_to_array(self, int i):
        cdef AllocationContainer *obj = &self.containers[i]
        cdef np.uint8_t[:] o = <np.uint8_t[:obj.n]> (<np.uint8_t*> obj.my_objs)
        rv = np.asarray(o)
        return rv



import abc
import weakref
from functools import cached_property
from numbers import Number
from typing import Optional, Tuple

import numpy as np

from yt._typing import AxisOrder
from yt.funcs import fix_unitary, is_sequence, parse_center_array, validate_width_tuple
from yt.units.yt_array import YTArray, YTQuantity
from yt.utilities.exceptions import YTCoordinateNotImplemented, YTInvalidWidthError


def _unknown_coord(field, data):
    raise YTCoordinateNotImplemented


def _get_coord_fields(axi, units="code_length"):
    def _dds(field, data):
        rv = data.ds.arr(data.fwidth[..., axi].copy(), units)
        return data._reshape_vals(rv)

    def _coords(field, data):
        rv = data.ds.arr(data.fcoords[..., axi].copy(), units)
        return data._reshape_vals(rv)

    return _dds, _coords


def _get_vert_fields(axi, units="code_length"):
    def _vert(field, data):
        rv = data.ds.arr(data.fcoords_vertex[..., axi].copy(), units)
        return rv

    return _vert


def _setup_dummy_cartesian_coords_and_widths(registry, axes: Tuple[str]):
    for ax in axes:
        registry.add_field(
            ("index", f"d{ax}"), sampling_type="cell", function=_unknown_coord
        )
        registry.add_field(("index", ax), sampling_type="cell", function=_unknown_coord)


def _setup_polar_coordinates(registry, axis_id):
    f1, f2 = _get_coord_fields(axis_id["r"])
    registry.add_field(
        ("index", "dr"),
        sampling_type="cell",
        function=f1,
        display_field=False,
        units="code_length",
    )

    registry.add_field(
        ("index", "r"),
        sampling_type="cell",
        function=f2,
        display_field=False,
        units="code_length",
    )

    f1, f2 = _get_coord_fields(axis_id["theta"], "dimensionless")
    registry.add_field(
        ("index", "dtheta"),
        sampling_type="cell",
        function=f1,
        display_field=False,
        units="dimensionless",
    )

    registry.add_field(
        ("index", "theta"),
        sampling_type="cell",
        function=f2,
        display_field=False,
        units="dimensionless",
    )

    def _path_r(field, data):
        return data["index", "dr"]

    registry.add_field(
        ("index", "path_element_r"),
        sampling_type="cell",
        function=_path_r,
        units="code_length",
    )

    def _path_theta(field, data):
        # Note: this already assumes cell-centered
        return data["index", "r"] * data["index", "dtheta"]

    registry.add_field(
        ("index", "path_element_theta"),
        sampling_type="cell",
        function=_path_theta,
        units="code_length",
    )


def validate_sequence_width(width, ds, unit=None):
    if isinstance(width[0], tuple) and isinstance(width[1], tuple):
        validate_width_tuple(width[0])
        validate_width_tuple(width[1])
        return (
            ds.quan(width[0][0], fix_unitary(width[0][1])),
            ds.quan(width[1][0], fix_unitary(width[1][1])),
        )
    elif isinstance(width[0], Number) and isinstance(width[1], Number):
        return (ds.quan(width[0], "code_length"), ds.quan(width[1], "code_length"))
    elif isinstance(width[0], YTQuantity) and isinstance(width[1], YTQuantity):
        return (ds.quan(width[0]), ds.quan(width[1]))
    else:
        validate_width_tuple(width)
        # If width and unit are both valid width tuples, we
        # assume width controls x and unit controls y
        try:
            validate_width_tuple(unit)
            return (
                ds.quan(width[0], fix_unitary(width[1])),
                ds.quan(unit[0], fix_unitary(unit[1])),
            )
        except YTInvalidWidthError:
            return (
                ds.quan(width[0], fix_unitary(width[1])),
                ds.quan(width[0], fix_unitary(width[1])),
            )


class CoordinateHandler(abc.ABC):
    name: str
    _default_axis_order: AxisOrder

    def __init__(self, ds, ordering: Optional[AxisOrder] = None):
        self.ds = weakref.proxy(ds)
        if ordering is not None:
            self.axis_order = ordering
        else:
            self.axis_order = self._default_axis_order

    @abc.abstractmethod
    def setup_fields(self):
        # This should return field definitions for x, y, z, r, theta, phi
        pass

    @abc.abstractmethod
    def pixelize(self, dimension, data_source, field, bounds, size, antialias=True):
        # This should *actually* be a pixelize call, not just returning the
        # pixelizer
        pass

    @abc.abstractmethod
    def pixelize_line(self, field, start_point, end_point, npoints):
        pass

    def distance(self, start, end):
        p1 = self.convert_to_cartesian(start)
        p2 = self.convert_to_cartesian(end)
        return np.sqrt(((p1 - p2) ** 2.0).sum())

    @abc.abstractmethod
    def convert_from_cartesian(self, coord):
        pass

    @abc.abstractmethod
    def convert_to_cartesian(self, coord):
        pass

    @abc.abstractmethod
    def convert_to_cylindrical(self, coord):
        pass

    @abc.abstractmethod
    def convert_from_cylindrical(self, coord):
        pass

    @abc.abstractmethod
    def convert_to_spherical(self, coord):
        pass

    @abc.abstractmethod
    def convert_from_spherical(self, coord):
        pass

    @cached_property
    def data_projection(self):
        return {ax: None for ax in self.axis_order}

    @cached_property
    def data_transform(self):
        return {ax: None for ax in self.axis_order}

    @cached_property
    def axis_name(self):
        an = {}
        for axi, ax in enumerate(self.axis_order):
            an[axi] = ax
            an[ax] = ax
            an[ax.capitalize()] = ax
        return an

    @cached_property
    def axis_id(self):
        ai = {}
        for axi, ax in enumerate(self.axis_order):
            ai[ax] = ai[axi] = axi
        return ai

    @property
    def image_axis_name(self):
        rv = {}
        for i in range(3):
            rv[i] = (self.axis_name[self.x_axis[i]], self.axis_name[self.y_axis[i]])
            rv[self.axis_name[i]] = rv[i]
            rv[self.axis_name[i].capitalize()] = rv[i]
        return rv

    @cached_property
    def x_axis(self):
        ai = self.axis_id
        xa = {}
        for a1, a2 in self._x_pairs:
            xa[a1] = xa[ai[a1]] = ai[a2]
        return xa

    @cached_property
    def y_axis(self):
        ai = self.axis_id
        ya = {}
        for a1, a2 in self._y_pairs:
            ya[a1] = ya[ai[a1]] = ai[a2]
        return ya

    @property
    @abc.abstractmethod
    def period(self):
        pass

    def sanitize_depth(self, depth):
        if is_sequence(depth):
            validate_width_tuple(depth)
            depth = (self.ds.quan(depth[0], fix_unitary(depth[1])),)
        elif isinstance(depth, Number):
            depth = (
                self.ds.quan(depth, "code_length", registry=self.ds.unit_registry),
            )
        elif isinstance(depth, YTQuantity):
            depth = (depth,)
        else:
            raise YTInvalidWidthError(depth)
        return depth

    def sanitize_width(self, axis, width, depth):
        if width is None:
            # initialize the index if it is not already initialized
            self.ds.index
            # Default to code units
            if not is_sequence(axis):
                xax = self.x_axis[axis]
                yax = self.y_axis[axis]
                w = self.ds.domain_width[np.array([xax, yax])]
            else:
                # axis is actually the normal vector
                # for an off-axis data object.
                mi = np.argmin(self.ds.domain_width)
                w = self.ds.domain_width[np.array((mi, mi))]
            width = (w[0], w[1])
        elif is_sequence(width):
            width = validate_sequence_width(width, self.ds)
        elif isinstance(width, YTQuantity):
            width = (width, width)
        elif isinstance(width, Number):
            width = (
                self.ds.quan(width, "code_length"),
                self.ds.quan(width, "code_length"),
            )
        else:
            raise YTInvalidWidthError(width)
        if depth is not None:
            depth = self.sanitize_depth(depth)
            return width + depth
        return width

    def sanitize_center(self, center, axis):
        center = parse_center_array(center, ds=self.ds, axis=axis)
        # This has to return both a center and a display_center
        display_center = self.convert_to_cartesian(center)
        return center, display_center


def cartesian_to_cylindrical(coord, center=(0, 0, 0)):
    c2 = np.zeros_like(coord)
    if not isinstance(center, YTArray):
        center = center * coord.uq
    c2[..., 0] = (
        (coord[..., 0] - center[0]) ** 2.0 + (coord[..., 1] - center[1]) ** 2.0
    ) ** 0.5
    c2[..., 1] = coord[..., 2]  # rzt
    c2[..., 2] = np.arctan2(coord[..., 1] - center[1], coord[..., 0] - center[0])
    return c2


def cylindrical_to_cartesian(coord, center=(0, 0, 0)):
    c2 = np.zeros_like(coord)
    if not isinstance(center, YTArray):
        center = center * coord.uq
    c2[..., 0] = np.cos(coord[..., 0]) * coord[..., 1] + center[0]
    c2[..., 1] = np.sin(coord[..., 0]) * coord[..., 1] + center[1]
    c2[..., 2] = coord[..., 2]
    return c2


def _get_polar_bounds(self: CoordinateHandler, axes: Tuple[str, str]):
    # a small helper function that is needed by two unrelated classes
    ri = self.axis_id[axes[0]]
    pi = self.axis_id[axes[1]]
    rmin = self.ds.domain_left_edge[ri]
    rmax = self.ds.domain_right_edge[ri]
    phimin = self.ds.domain_left_edge[pi]
    phimax = self.ds.domain_right_edge[pi]
    corners = [
        (rmin, phimin),
        (rmin, phimax),
        (rmax, phimin),
        (rmax, phimax),
    ]

    def to_polar_plane(r, phi):
        x = r * np.cos(phi)
        y = r * np.sin(phi)
        return x, y

    conic_corner_coords = [to_polar_plane(*corner) for corner in corners]

    phimin = phimin.d
    phimax = phimax.d

    if phimin <= np.pi <= phimax:
        xxmin = -rmax
    else:
        xxmin = min(xx for xx, yy in conic_corner_coords)

    if phimin <= 0 <= phimax:
        xxmax = rmax
    else:
        xxmax = max(xx for xx, yy in conic_corner_coords)

    if phimin <= 3 * np.pi / 2 <= phimax:
        yymin = -rmax
    else:
        yymin = min(yy for xx, yy in conic_corner_coords)

    if phimin <= np.pi / 2 <= phimax:
        yymax = rmax
    else:
        yymax = max(yy for xx, yy in conic_corner_coords)

    return xxmin, xxmax, yymin, yymax



from yt.loaders import load

DEFAULT_ATTRS = (
    "dimensionality",
    "refine_by",
    "domain_dimensions",
    "current_time",
    "domain_left_edge",
    "domain_right_edge",
    "unique_identifier",
    "current_redshift",
    "cosmological_simulation",
    "omega_matter",
    "omega_lambda",
    "hubble_constant",
    "dataset_type",
)


def get_metadata(path, full_output=False, attrs=DEFAULT_ATTRS):
    ds = load(path)
    metadata = {"filename": path}
    for a in attrs:
        v = getattr(ds, a, None)
        if v is None:
            continue
        if hasattr(v, "tolist"):
            v = v.tolist()
        metadata[a] = v
    if full_output:
        params = {}
        for p, v in ds.parameters.items():
            if hasattr(v, "tolist"):
                v = v.tolist()
            params[p] = v
        metadata["params"] = params
    ds.close()
    return metadata



from .oct_visitors cimport CountTotalCells, CountTotalOcts


cdef class SelectorObject:

    def __cinit__(self, dobj, *args):
        self._hash_initialized = 0
        cdef np.float64_t [:] DLE
        cdef np.float64_t [:] DRE
        min_level = getattr(dobj, "min_level", None)
        max_level = getattr(dobj, "max_level", None)
        if min_level is None:
            min_level = 0
        if max_level is None:
            max_level = 99
        self.min_level = min_level
        self.max_level = max_level
        self.overlap_cells = 0

        ds = getattr(dobj, 'ds', None)
        if ds is None:
            for i in range(3):
                # NOTE that this is not universal.
                self.domain_width[i] = 1.0
                self.periodicity[i] = False
        else:
            DLE = _ensure_code(ds.domain_left_edge)
            DRE = _ensure_code(ds.domain_right_edge)
            for i in range(3):
                self.domain_width[i] = DRE[i] - DLE[i]
                self.domain_center[i] = DLE[i] + 0.5 * self.domain_width[i]
                self.periodicity[i] = ds.periodicity[i]

    def get_periodicity(self):
        cdef int i
        cdef np.ndarray[np.uint8_t, ndim=1] periodicity
        periodicity = np.zeros(3, dtype='uint8')
        for i in range(3):
            periodicity[i] = self.periodicity[i]
        return periodicity

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def select_grids(self,
                     np.ndarray[np.float64_t, ndim=2] left_edges,
                     np.ndarray[np.float64_t, ndim=2] right_edges,
                     np.ndarray[np.int32_t, ndim=2] levels):
        cdef int i, n
        cdef int ng = left_edges.shape[0]
        cdef np.ndarray[np.uint8_t, ndim=1] gridi = np.zeros(ng, dtype='uint8')
        cdef np.float64_t LE[3]
        cdef np.float64_t RE[3]
        _ensure_code(left_edges)
        _ensure_code(right_edges)
        with nogil:
            for n in range(ng):
                # Call our selector function
                # Check if the sphere is inside the grid
                for i in range(3):
                    LE[i] = left_edges[n, i]
                    RE[i] = right_edges[n, i]
                gridi[n] = self.select_grid(LE, RE, levels[n, 0])
        return gridi.astype("bool")

    def count_octs(self, OctreeContainer octree, int domain_id = -1):
        cdef CountTotalOcts visitor
        visitor = CountTotalOcts(octree, domain_id)
        octree.visit_all_octs(self, visitor)
        return visitor.index

    def count_oct_cells(self, OctreeContainer octree, int domain_id = -1):
        cdef CountTotalCells visitor
        visitor = CountTotalCells(octree, domain_id)
        octree.visit_all_octs(self, visitor)
        return visitor.index

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef void recursively_visit_octs(self, Oct *root,
                        np.float64_t pos[3], np.float64_t dds[3],
                        int level,
                        OctVisitor visitor,
                        int visit_covered = 0):
        # visit_covered tells us whether this octree supports partial
        # refinement.  If it does, we need to handle this specially -- first
        # we visit *this* oct, then we make a second pass to check any child
        # octs.
        cdef np.float64_t LE[3]
        cdef np.float64_t RE[3]
        cdef np.float64_t sdds[3]
        cdef np.float64_t spos[3]
        cdef int i, j, k, res
        cdef Oct *ch
        # Remember that pos is the *center* of the oct, and dds is the oct
        # width.  So to get to the edges, we add/subtract half of dds.
        for i in range(3):
            # sdds is the cell width
            sdds[i] = dds[i]/2.0
            LE[i] = pos[i] - dds[i]/2.0
            RE[i] = pos[i] + dds[i]/2.0
        #print(LE[0], RE[0], LE[1], RE[1], LE[2], RE[2])
        res = self.select_grid(LE, RE, level, root)
        if res == 1 and visitor.domain > 0 and root.domain != visitor.domain:
            res = -1
        cdef int increment = 1
        cdef int next_level, this_level
        # next_level: an int that says whether or not we can progress to children
        # this_level: an int that says whether or not we can select from this
        # level
        next_level = this_level = 1
        if res == -1:
            # This happens when we do domain selection but the oct has
            # children.  This would allow an oct to pass to its children but
            # not get accessed itself.
            next_level = 1
            this_level = 0
        elif level == self.max_level:
            next_level = 0
        elif level < self.min_level or level > self.max_level:
            this_level = 0
        if res == 0 and this_level == 1:
            return
        # Now we visit all our children.  We subtract off sdds for the first
        # pass because we center it on the first cell.
        cdef int iter = 1 - visit_covered # 2 if 1, 1 if 0.
        # So the order here goes like so.  If visit_covered is 1, which usually
        # comes from "partial_coverage", we visit the components of a zone even
        # if it has children.  But in general, the first iteration through, we
        # visit each cell.  This means that only if visit_covered is true do we
        # visit potentially covered cells.  The next time through, we visit
        # child cells.
        while iter < 2:
            spos[0] = pos[0] - sdds[0]/2.0
            for i in range(2):
                spos[1] = pos[1] - sdds[1]/2.0
                for j in range(2):
                    spos[2] = pos[2] - sdds[2]/2.0
                    for k in range(2):
                        ch = NULL
                        # We only supply a child if we are actually going to
                        # look at the next level.
                        if root.children != NULL and next_level == 1:
                            ch = root.children[cind(i, j, k)]
                        if iter == 1 and next_level == 1 and ch != NULL:
                            # Note that visitor.pos is always going to be the
                            # position of the Oct -- it is *not* always going
                            # to be the same as the position of the cell under
                            # investigation.
                            visitor.pos[0] = (visitor.pos[0] << 1) + i
                            visitor.pos[1] = (visitor.pos[1] << 1) + j
                            visitor.pos[2] = (visitor.pos[2] << 1) + k
                            visitor.level += 1
                            self.recursively_visit_octs(
                                ch, spos, sdds, level + 1, visitor,
                                visit_covered)
                            visitor.pos[0] = (visitor.pos[0] >> 1)
                            visitor.pos[1] = (visitor.pos[1] >> 1)
                            visitor.pos[2] = (visitor.pos[2] >> 1)
                            visitor.level -= 1
                        elif this_level == 1 and visitor.nz > 1:
                            visitor.global_index += increment
                            increment = 0
                            self.visit_oct_cells(root, ch, spos, sdds,
                                                 visitor, i, j, k)
                        elif this_level == 1 and increment == 1:
                            visitor.global_index += increment
                            increment = 0
                            visitor.ind[0] = visitor.ind[1] = visitor.ind[2] = 0
                            visitor.visit(root, 1)
                        spos[2] += sdds[2]
                    spos[1] += sdds[1]
                spos[0] += sdds[0]
            this_level = 0 # We turn this off for the second pass.
            iter += 1

    cdef void visit_oct_cells(self, Oct *root, Oct *ch,
                              np.float64_t spos[3], np.float64_t sdds[3],
                              OctVisitor visitor, int i, int j, int k):
        # We can short-circuit the whole process if data.nz == 2.
        # This saves us some funny-business.
        cdef int selected
        if visitor.nz == 2:
            selected = self.select_cell(spos, sdds)
            if ch != NULL:
                selected *= self.overlap_cells
            # visitor.ind refers to the cell, not to the oct.
            visitor.ind[0] = i
            visitor.ind[1] = j
            visitor.ind[2] = k
            visitor.visit(root, selected)
            return
        # Okay, now that we've got that out of the way, we have to do some
        # other checks here.  In this case, spos[] is the position of the
        # center of a *possible* oct child, which means it is the center of a
        # cluster of cells.  That cluster might have 1, 8, 64, ... cells in it.
        # But, we can figure it out by calculating the cell dds.
        cdef np.float64_t dds[3]
        cdef np.float64_t pos[3]
        cdef int ci, cj, ck
        cdef int nr = (visitor.nz >> 1)
        for ci in range(3):
            dds[ci] = sdds[ci] / nr
        # Boot strap at the first index.
        pos[0] = (spos[0] - sdds[0]/2.0) + dds[0] * 0.5
        for ci in range(nr):
            pos[1] = (spos[1] - sdds[1]/2.0) + dds[1] * 0.5
            for cj in range(nr):
                pos[2] = (spos[2] - sdds[2]/2.0) + dds[2] * 0.5
                for ck in range(nr):
                    selected = self.select_cell(pos, dds)
                    if ch != NULL:
                        selected *= self.overlap_cells
                    visitor.ind[0] = ci + i * nr
                    visitor.ind[1] = cj + j * nr
                    visitor.ind[2] = ck + k * nr
                    visitor.visit(root, selected)
                    pos[2] += dds[2]
                pos[1] += dds[1]
            pos[0] += dds[0]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_grid(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3],
                               np.int32_t level, Oct *o = NULL) noexcept nogil:
        if level < self.min_level or level > self.max_level: return 0
        return self.select_bbox(left_edge, right_edge)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_grid_edge(self, np.float64_t left_edge[3],
                                    np.float64_t right_edge[3],
                                    np.int32_t level, Oct *o = NULL) noexcept nogil:
        if level < self.min_level or level > self.max_level: return 0
        return self.select_bbox_edge(left_edge, right_edge)

    cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) noexcept nogil:
        return 0

    cdef int select_point(self, np.float64_t pos[3]) noexcept nogil:
        return 0

    cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) noexcept nogil:
        return 0

    cdef int select_bbox(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) noexcept nogil:
        """
        Returns:
          0: If the selector does not touch the bounding box.
          1: If the selector overlaps the bounding box anywhere.
        """
        return 0

    cdef int select_bbox_edge(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) noexcept nogil:
        """
        Returns:
          0: If the selector does not touch the bounding box.
          1: If the selector contains the entire bounding box.
          2: If the selector contains part of the bounding box.
        """
        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef np.float64_t periodic_difference(self, np.float64_t x1, np.float64_t x2, int d) nogil:
        # domain_width is already in code units, and we assume what is fed in
        # is too.
        cdef np.float64_t rel = x1 - x2
        if self.periodicity[d]:
            if rel > self.domain_width[d] * 0.5:
                rel -= self.domain_width[d]
            elif rel < -self.domain_width[d] * 0.5:
                rel += self.domain_width[d]
        return rel

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def fill_mesh_mask(self, mesh):
        cdef np.float64_t pos[3]
        cdef np.ndarray[np.int64_t, ndim=2] indices
        cdef np.ndarray[np.float64_t, ndim=2] coords
        cdef np.ndarray[np.uint8_t, ndim=1] mask
        cdef int i, j, k, selected
        cdef int npoints, nv = mesh._connectivity_length
        cdef int total = 0
        cdef int offset = mesh._index_offset
        coords = _ensure_code(mesh.connectivity_coords)
        indices = mesh.connectivity_indices
        npoints = indices.shape[0]
        mask = np.zeros(npoints, dtype='uint8')
        for i in range(npoints):
            selected = 0
            for j in range(nv):
                for k in range(3):
                    pos[k] = coords[indices[i, j] - offset, k]
                selected = self.select_point(pos)
                if selected == 1: break
            total += selected
            mask[i] = selected
        if total == 0: return None
        return mask.astype("bool")

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def fill_mesh_cell_mask(self, mesh):
        cdef np.float64_t pos
        cdef np.float64_t le[3]
        cdef np.float64_t re[3]
        cdef np.ndarray[np.int64_t, ndim=2] indices
        cdef np.ndarray[np.float64_t, ndim=2] coords
        cdef np.ndarray[np.uint8_t, ndim=1] mask
        cdef int i, j, k, selected
        cdef int npoints, nv = mesh._connectivity_length
        cdef int ndim = mesh.connectivity_coords.shape[1]
        cdef int total = 0
        cdef int offset = mesh._index_offset
        coords = _ensure_code(mesh.connectivity_coords)
        indices = mesh.connectivity_indices
        npoints = indices.shape[0]
        mask = np.zeros(npoints, dtype='uint8')
        for i in range(npoints):
            selected = 0
            for k in range(3):
                le[k] = 1e60
                re[k] = -1e60
            for j in range(nv):
                for k in range(ndim):
                    pos = coords[indices[i, j] - offset, k]
                    le[k] = fmin(pos, le[k])
                    re[k] = fmax(pos, re[k])
                for k in range(2, ndim - 1, -1):
                    le[k] = self.domain_center[k]
                    re[k] = self.domain_center[k]
            selected = self.select_bbox(le, re)
            total += selected
            mask[i] = selected
        if total == 0: return None
        return mask.astype("bool")

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def fill_mask_regular_grid(self, gobj):
        cdef np.ndarray[np.uint8_t, ndim=3, cast=True] child_mask
        child_mask = gobj.child_mask
        cdef np.ndarray[np.uint8_t, ndim=3] mask
        cdef int dim[3]
        _ensure_code(gobj.dds)
        _ensure_code(gobj.LeftEdge)
        _ensure_code(gobj.RightEdge)
        cdef np.ndarray[np.float64_t, ndim=1] odds = gobj.dds.d
        cdef np.ndarray[np.float64_t, ndim=1] oleft_edge = gobj.LeftEdge.d
        cdef np.ndarray[np.float64_t, ndim=1] oright_edge = gobj.RightEdge.d
        cdef int i
        cdef np.float64_t dds[3]
        cdef np.float64_t left_edge[3]
        cdef np.float64_t right_edge[3]
        for i in range(3):
            dds[i] = odds[i]
            dim[i] = gobj.ActiveDimensions[i]
            left_edge[i] = oleft_edge[i]
            right_edge[i] = oright_edge[i]
        mask = np.zeros(gobj.ActiveDimensions, dtype='uint8')
        # Check for the level bounds
        cdef np.int32_t level = gobj.Level
        # We set this to 1 if we ignore child_mask
        cdef int total
        total = self.fill_mask_selector_regular_grid(left_edge, right_edge,
                                                     dds, dim, child_mask,
                                                     mask, level)
        if total == 0: return None
        return mask.astype("bool")

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int fill_mask_selector_regular_grid(self, np.float64_t left_edge[3],
                                np.float64_t right_edge[3],
                                np.float64_t dds[3], int dim[3],
                                np.ndarray[np.uint8_t, ndim=3, cast=True] child_mask,
                                np.ndarray[np.uint8_t, ndim=3] mask,
                                int level):
        cdef int i, j, k
        cdef int total = 0, this_level = 0
        cdef np.float64_t pos[3]
        if level < self.min_level or level > self.max_level:
            return 0
        if level == self.max_level:
            this_level = 1
        with nogil:
            for i in range(dim[0]):
                pos[0] = left_edge[0] + (i + 0.5) * dds[0]
                for j in range(dim[1]):
                    pos[1] = left_edge[1] + (j + 0.5) * dds[1]
                    for k in range(dim[2]):
                        pos[2] = left_edge[2] + (k + 0.5) * dds[2]
                        if child_mask[i, j, k] == 1 or this_level == 1:
                            mask[i, j, k] = self.select_cell(pos, dds)
                            total += mask[i, j, k]
        return total

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def fill_mask(self, gobj):
        # This is for an irregular grid.  We make no assumptions about the
        # shape of the dds values, which are supplied as differing-length
        # arrays.
        cdef np.ndarray[np.uint8_t, ndim=3, cast=True] child_mask
        child_mask = gobj.child_mask
        cdef np.ndarray[np.uint8_t, ndim=3] mask
        cdef int dim[3]
        _ensure_code(gobj.cell_widths[0])
        _ensure_code(gobj.cell_widths[1])
        _ensure_code(gobj.cell_widths[2])
        _ensure_code(gobj.LeftEdge)
        _ensure_code(gobj.RightEdge)
        cdef np.ndarray[np.float64_t, ndim=1] oleft_edge = gobj.LeftEdge.d
        cdef np.ndarray[np.float64_t, ndim=1] oright_edge = gobj.RightEdge.d
        cdef np.ndarray[np.float64_t, ndim=1] ocell_width
        cdef int i, n = 0
        cdef np.float64_t left_edge[3]
        cdef np.float64_t right_edge[3]
        cdef np.float64_t **dds
        dds = <np.float64_t**> malloc(sizeof(np.float64_t*)*3)
        for i in range(3):
            dim[i] = gobj.ActiveDimensions[i]
            left_edge[i] = oleft_edge[i]
            right_edge[i] = oright_edge[i]
            dds[i] = <np.float64_t *> malloc(sizeof(np.float64_t) * dim[i])
            ocell_width = gobj.cell_widths[i]
            for j in range(dim[i]):
                dds[i][j] = ocell_width[j]
        mask = np.zeros(gobj.ActiveDimensions, dtype='uint8')
        # Check for the level bounds
        cdef np.int32_t level = gobj.Level
        # We set this to 1 if we ignore child_mask
        cdef int total
        total = self.fill_mask_selector(left_edge, right_edge,
                                        dds, dim, child_mask,
                                        mask, level)
        for i in range(3):
            free(dds[i])
        free(dds)
        if total == 0: return None
        return mask.astype("bool")

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int fill_mask_selector(self, np.float64_t left_edge[3],
                                np.float64_t right_edge[3],
                                np.float64_t **dds, int dim[3],
                                np.ndarray[np.uint8_t, ndim=3, cast=True] child_mask,
                                np.ndarray[np.uint8_t, ndim=3] mask,
                                int level):
        cdef int i, j, k
        cdef int total = 0, this_level = 0
        cdef np.float64_t pos[3]
        if level < self.min_level or level > self.max_level:
            return 0
        if level == self.max_level:
            this_level = 1
        cdef np.float64_t *offsets[3]
        cdef np.float64_t c = 0.0
        cdef np.float64_t tdds[3]
        for i in range(3):
            offsets[i] = <np.float64_t *> malloc(dim[i] * sizeof(np.float64_t))
            c = left_edge[i]
            for j in range(dim[i]):
                offsets[i][j] = c
                c += dds[i][j]
        with nogil:
            # We need to keep in mind that it is entirely possible to
            # accumulate round-off error by doing lots of additions, etc.
            # That's one of the reasons we construct (ahead of time) the edge
            # array. I mean, we don't necessarily *have* to do that, but it
            # seems OK.
            for i in range(dim[0]):
                tdds[0] = dds[0][i]
                pos[0] = offsets[0][i] + 0.5 * tdds[0]
                for j in range(dim[1]):
                    tdds[1] = dds[1][j]
                    pos[1] = offsets[1][j] + 0.5 * tdds[1]
                    for k in range(dim[2]):
                        tdds[2] = dds[2][k]
                        pos[2] = offsets[2][k] + 0.5 * tdds[2]
                        if child_mask[i, j, k] == 1 or this_level == 1:
                            mask[i, j, k] = self.select_cell(pos, tdds)
                            total += mask[i, j, k]
        free(offsets[0])
        free(offsets[1])
        free(offsets[2])
        return total

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef void visit_grid_cells(self, GridVisitorData *data,
                              grid_visitor_function *func,
                              np.uint8_t *cached_mask = NULL):
        # This function accepts a grid visitor function, the data that
        # corresponds to the current grid being examined (the most important
        # aspect of which is the .grid attribute, along with index values and
        # void* pointers to arrays) and a possibly-pre-generated cached mask.
        # Each cell is visited with the grid visitor function.
        cdef np.float64_t left_edge[3]
        cdef np.float64_t right_edge[3]
        cdef np.float64_t dds[3]
        cdef int dim[3]
        cdef int this_level = 0, level, i
        cdef np.float64_t pos[3]
        level = data.grid.level
        if level < self.min_level or level > self.max_level:
            return
        if level == self.max_level:
            this_level = 1
        cdef np.uint8_t child_masked, selected
        for i in range(3):
            left_edge[i] = data.grid.left_edge[i]
            right_edge[i] = data.grid.right_edge[i]
            dds[i] = (right_edge[i] - left_edge[i])/data.grid.dims[i]
            dim[i] = data.grid.dims[i]
        with nogil:
            pos[0] = left_edge[0] + dds[0] * 0.5
            data.pos[0] = 0
            for i in range(dim[0]):
                pos[1] = left_edge[1] + dds[1] * 0.5
                data.pos[1] = 0
                for j in range(dim[1]):
                    pos[2] = left_edge[2] + dds[2] * 0.5
                    data.pos[2] = 0
                    for k in range(dim[2]):
                        # We short-circuit if we have a cache; if we don't, we
                        # only set selected to true if it's *not* masked by a
                        # child and it *is* selected.
                        if cached_mask != NULL:
                            selected = ba_get_value(cached_mask,
                                                    data.global_index)
                        else:
                            if this_level == 1:
                                child_masked = 0
                            else:
                                child_masked = check_child_masked(data)
                            if child_masked == 0:
                                selected = self.select_cell(pos, dds)
                            else:
                                selected = 0
                        func(data, selected)
                        data.global_index += 1
                        pos[2] += dds[2]
                        data.pos[2] += 1
                    pos[1] += dds[1]
                    data.pos[1] += 1
                pos[0] += dds[0]
                data.pos[0] += 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def count_points(self, np.ndarray[cython.floating, ndim=1] x,
                           np.ndarray[cython.floating, ndim=1] y,
                           np.ndarray[cython.floating, ndim=1] z,
                           radii):
        cdef int count = 0
        cdef int i
        cdef np.float64_t pos[3]
        cdef np.float64_t radius
        cdef np.float64_t[:] _radii
        if radii is not None:
            _radii = np.atleast_1d(np.array(radii, dtype='float64'))
        else:
            _radii = np.array([0.0], dtype='float64')
        _ensure_code(x)
        _ensure_code(y)
        _ensure_code(z)
        with nogil:
            for i in range(x.shape[0]):
                pos[0] = x[i]
                pos[1] = y[i]
                pos[2] = z[i]
                if _radii.shape[0] == 1:
                    radius = _radii[0]
                else:
                    radius = _radii[i]
                if radius == 0:
                    count += self.select_point(pos)
                else:
                    count += self.select_sphere(pos, radius)
        return count

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def select_points(self,
                      np.ndarray[cython.floating, ndim=1] x,
                      np.ndarray[cython.floating, ndim=1] y,
                      np.ndarray[cython.floating, ndim=1] z,
                      radii):
        cdef int count = 0
        cdef int i
        cdef np.float64_t pos[3]
        cdef np.float64_t radius
        cdef np.ndarray[np.uint8_t, ndim=1] mask
        cdef np.float64_t[:] _radii
        if radii is not None:
            _radii = np.atleast_1d(np.array(radii, dtype='float64'))
        else:
            _radii = np.array([0.0], dtype='float64')
        mask = np.empty(x.shape[0], dtype='uint8')
        _ensure_code(x)
        _ensure_code(y)
        _ensure_code(z)


        # this is to allow selectors to optimize the point vs
        # 0-radius sphere case.  These two may have different
        # effects for 0-volume selectors, however (collision
        # between a ray and a point is null, while ray and a
        # sphere is allowed)
        with nogil:
            for i in range(x.shape[0]) :
                pos[0] = x[i]
                pos[1] = y[i]
                pos[2] = z[i]
                if _radii.shape[0] == 1:
                    radius = 0
                else:
                    radius = _radii[i]
                if radius == 0:
                    mask[i] = self.select_point(pos)
                else:
                    mask[i] = self.select_sphere(pos, radius)
                count += mask[i]
        if count == 0: return None
        return mask.view("bool")

    def __hash__(self):
        # convert data to be hashed to a byte array, which FNV algorithm expects
        if self._hash_initialized == 1:
            return self._hash
        hash_data = bytearray()
        for v in self._hash_vals() + self._base_hash():
            if isinstance(v, tuple):
                hash_data.extend(v[0].encode('ascii'))
                hash_data.extend(repr(v[1]).encode('ascii'))
            else:
                hash_data.extend(repr(v).encode('ascii'))
        cdef np.int64_t hash_value = fnv_hash(hash_data)
        self._hash = hash_value
        self._hash_initialized = 1
        return hash_value

    def _hash_vals(self):
        raise NotImplementedError

    def _base_hash(self):
        return (("min_level", self.min_level),
                ("max_level", self.max_level),
                ("overlap_cells", self.overlap_cells),
                ("periodicity[0]", self.periodicity[0]),
                ("periodicity[1]", self.periodicity[1]),
                ("periodicity[2]", self.periodicity[2]),
                ("domain_width[0]", self.domain_width[0]),
                ("domain_width[1]", self.domain_width[1]),
                ("domain_width[2]", self.domain_width[2]))

    def _get_state_attnames(self):
        # return a tupe of attr names for __setstate__: implement for each subclass
        raise NotImplementedError

    def __getstate__(self):
        # returns a tuple containing (attribute name, attribute value) tuples needed to
        # rebuild the state:
        base_atts = ("min_level", "max_level", "overlap_cells",
                     "periodicity", "domain_width", "domain_center")
        child_atts = self._get_state_attnames()

        # assemble the state_tuple (('a1', a1val), ('a2', a2val),...)
        state_tuple = ()
        for fld in base_atts + child_atts:
            state_tuple += ((fld, getattr(self, fld)), )
        return state_tuple

    def __getnewargs__(self):
        # __setstate__ will always call __cinit__, this pickle hook returns arguments
        # to __cinit__. We will give it None so we dont error then set attributes in
        # __setstate__ Note that we could avoid this by making dobj an optional argument
        # to __cinit__
        return (None, )

    def __setstate__(self, state_tuple):
        # parse and set attributes from the state_tuple: (('a1',a1val),('a2',a2val),...)
        for attr in state_tuple:
            setattr(self, attr[0], attr[1])



import numpy as np
from numpy.testing import assert_almost_equal, assert_array_equal, assert_equal

from yt.fields.derived_field import ValidateParameter
from yt.loaders import load, load_particles
from yt.testing import (
    fake_octree_ds,
    fake_random_ds,
    requires_file,
    requires_module,
)
from yt.units import kpc

# cylindrical data for covering_grid test
cyl_2d = "WDMerger_hdf5_chk_1000/WDMerger_hdf5_chk_1000.hdf5"
cyl_3d = "MHD_Cyl3d_hdf5_plt_cnt_0100/MHD_Cyl3d_hdf5_plt_cnt_0100.hdf5"


def setup():
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


@requires_module("h5py")
@requires_file(cyl_2d)
@requires_file(cyl_3d)
def test_covering_grid():
    # We decompose in different ways
    for level in [0, 1, 2]:
        for nprocs in [1, 2, 4, 8]:
            ds = fake_random_ds(16, nprocs=nprocs)
            axis_name = ds.coordinates.axis_name
            dn = ds.refine_by**level
            cg = ds.covering_grid(level, [0.0, 0.0, 0.0], dn * ds.domain_dimensions)
            # Test coordinate generation
            assert_equal(np.unique(cg[("index", f"d{axis_name[0]}")]).size, 1)
            xmi = cg[("index", axis_name[0])].min()
            xma = cg[("index", axis_name[0])].max()
            dx = cg[("index", f"d{axis_name[0]}")].flat[0:1]
            edges = ds.arr([[0, 1], [0, 1], [0, 1]], "code_length")
            assert_equal(xmi, edges[0, 0] + dx / 2.0)
            assert_equal(xmi, cg[("index", axis_name[0])][0, 0, 0])
            assert_equal(xmi, cg[("index", axis_name[0])][0, 1, 1])
            assert_equal(xma, edges[0, 1] - dx / 2.0)
            assert_equal(xma, cg[("index", axis_name[0])][-1, 0, 0])
            assert_equal(xma, cg[("index", axis_name[0])][-1, 1, 1])
            assert_equal(np.unique(cg[("index", f"d{axis_name[1]}")]).size, 1)
            ymi = cg[("index", axis_name[1])].min()
            yma = cg[("index", axis_name[1])].max()
            dy = cg[("index", f"d{axis_name[1]}")][0]
            assert_equal(ymi, edges[1, 0] + dy / 2.0)
            assert_equal(ymi, cg[("index", axis_name[1])][0, 0, 0])
            assert_equal(ymi, cg[("index", axis_name[1])][1, 0, 1])
            assert_equal(yma, edges[1, 1] - dy / 2.0)
            assert_equal(yma, cg[("index", axis_name[1])][0, -1, 0])
            assert_equal(yma, cg[("index", axis_name[1])][1, -1, 1])
            assert_equal(np.unique(cg[("index", f"d{axis_name[2]}")]).size, 1)
            zmi = cg[("index", axis_name[2])].min()
            zma = cg[("index", axis_name[2])].max()
            dz = cg[("index", f"d{axis_name[2]}")][0]
            assert_equal(zmi, edges[2, 0] + dz / 2.0)
            assert_equal(zmi, cg[("index", axis_name[2])][0, 0, 0])
            assert_equal(zmi, cg[("index", axis_name[2])][1, 1, 0])
            assert_equal(zma, edges[2, 1] - dz / 2.0)
            assert_equal(zma, cg[("index", axis_name[2])][0, 0, -1])
            assert_equal(zma, cg[("index", axis_name[2])][1, 1, -1])
            # Now we test other attributes
            assert_equal(cg[("index", "ones")].max(), 1.0)
            assert_equal(cg[("index", "ones")].min(), 1.0)
            assert_equal(cg[("index", "grid_level")], level)
            assert_equal(cg[("index", "cell_volume")].sum(), ds.domain_width.prod())
            for g in ds.index.grids:
                di = g.get_global_startindex()
                dd = g.ActiveDimensions
                for i in range(dn):
                    f = cg[("gas", "density")][
                        dn * di[0] + i : dn * (di[0] + dd[0]) + i : dn,
                        dn * di[1] + i : dn * (di[1] + dd[1]) + i : dn,
                        dn * di[2] + i : dn * (di[2] + dd[2]) + i : dn,
                    ]
                    assert_equal(f, g[("gas", "density")])

    # More tests for cylindrical geometry
    for fn in [cyl_2d, cyl_3d]:
        ds = load(fn)
        ad = ds.all_data()
        upper_ad = ad.cut_region(["obj[('index', 'z')] > 0"])
        sp = ds.sphere((0, 0, 0), 0.5 * ds.domain_width[0], data_source=upper_ad)
        sp.quantities.total_mass()


@requires_module("h5py")
@requires_file(cyl_2d)
@requires_file(cyl_3d)
def test_covering_grid_data_source():
    # test the data_source kwarg (new with PR 4063)

    for level in [0, 1, 2]:
        for nprocs in [1, 2, 4, 8]:
            ds = fake_random_ds(16, nprocs=nprocs)
            dn = ds.refine_by**level
            cg = ds.covering_grid(level, [0.0, 0.0, 0.0], dn * ds.domain_dimensions)

            # check that the regularized sphere is centered where it should be,
            # to within the tolerance of the grid. Use a center offsect from
            # the domain center by a bit
            center = ds.domain_center + np.min(ds.domain_width) * 0.15
            sp = ds.sphere(center, (0.1, "code_length"))
            dims = dn * ds.domain_dimensions
            cg_sp = ds.covering_grid(level, [0.0, 0.0, 0.0], dims, data_source=sp)

            # find the discrete center of the sphere by averaging the position
            # along each dimension where values are non-zero
            cg_mask = cg_sp["gas", "density"] != 0
            discrete_c = ds.arr(
                [
                    np.mean(cg_sp[("index", dim)][cg_mask])
                    for dim in ds.coordinates.axis_order
                ]
            )

            # should be centered to within the tolerance of the grid at least
            grid_tol = ds.arr(
                [cg_sp[("index", dim)][0, 0, 0] for dim in ds.coordinates.axis_order]
            )
            assert np.all(np.abs(discrete_c - center) <= grid_tol)

            # check that a region covering the whole domain matches no data_source
            reg = ds.region(ds.domain_center, ds.domain_left_edge, ds.domain_right_edge)
            cg_reg = ds.covering_grid(level, [0.0, 0.0, 0.0], dims, data_source=reg)
            assert np.all(cg[("gas", "density")] == cg_reg[("gas", "density")])

            # check that a box covering a subset of the domain is the right volume
            right_edge = ds.domain_left_edge + ds.domain_width * 0.5
            c = (right_edge + ds.domain_left_edge) / 2
            reg = ds.region(c, ds.domain_left_edge, right_edge)
            cg_reg = ds.covering_grid(level, [0.0, 0.0, 0.0], dims, data_source=reg)
            box_vol = cg_reg[("index", "cell_volume")][
                cg_reg["gas", "density"] != 0
            ].sum()
            actual_vol = np.prod(right_edge - ds.domain_left_edge)
            assert box_vol == actual_vol


@requires_module("xarray")
def test_xarray_export():
    def _run_tests(cg):
        xarr = cg.to_xarray(fields=[("gas", "density"), ("gas", "temperature")])
        assert ("gas", "density") in xarr.variables
        assert ("gas", "temperature") in xarr.variables
        assert ("gas", "specific_thermal_energy") not in xarr.variables
        assert "x" in xarr.coords
        assert "y" in xarr.coords
        assert "z" in xarr.coords
        assert xarr.dims["x"] == dn * ds.domain_dimensions[0]
        assert xarr.dims["y"] == dn * ds.domain_dimensions[1]
        assert xarr.dims["z"] == dn * ds.domain_dimensions[2]
        assert_equal(xarr.x, cg[("index", "x")][:, 0, 0])
        assert_equal(xarr.y, cg[("index", "y")][0, :, 0])
        assert_equal(xarr.z, cg[("index", "z")][0, 0, :])

    fields = ("density", "temperature", "specific_thermal_energy")
    units = ("g/cm**3", "K", "erg/g")
    for level in [0, 1, 2]:
        ds = fake_random_ds(16, fields=fields, units=units)
        dn = ds.refine_by**level
        rcg = ds.covering_grid(level, [0.0, 0.0, 0.0], dn * ds.domain_dimensions)
        _run_tests(rcg)
        scg = ds.smoothed_covering_grid(
            level, [0.0, 0.0, 0.0], dn * ds.domain_dimensions
        )
        _run_tests(scg)
        ag1 = ds.arbitrary_grid(
            [0.0, 0.0, 0.0], [1.0, 1.0, 1.0], dn * ds.domain_dimensions
        )
        _run_tests(ag1)
        ag2 = ds.arbitrary_grid(
            [0.1, 0.3, 0.2], [0.4, 1.0, 0.9], dn * ds.domain_dimensions
        )
        _run_tests(ag2)


def test_smoothed_covering_grid():
    # We decompose in different ways
    for level in [0, 1, 2]:
        for nprocs in [1, 2, 4, 8]:
            ds = fake_random_ds(16, nprocs=nprocs)
            dn = ds.refine_by**level
            cg = ds.smoothed_covering_grid(
                level, [0.0, 0.0, 0.0], dn * ds.domain_dimensions
            )
            assert_equal(cg[("index", "ones")].max(), 1.0)
            assert_equal(cg[("index", "ones")].min(), 1.0)
            assert_equal(cg[("index", "cell_volume")].sum(), ds.domain_width.prod())
            for g in ds.index.grids:
                if level != g.Level:
                    continue
                di = g.get_global_startindex()
                dd = g.ActiveDimensions
                for i in range(dn):
                    f = cg[("gas", "density")][
                        dn * di[0] + i : dn * (di[0] + dd[0]) + i : dn,
                        dn * di[1] + i : dn * (di[1] + dd[1]) + i : dn,
                        dn * di[2] + i : dn * (di[2] + dd[2]) + i : dn,
                    ]
                    assert_equal(f, g[("gas", "density")])


def test_arbitrary_grid():
    for ncells in [32, 64]:
        for px in [0.125, 0.25, 0.55519]:
            particle_data = {
                "particle_position_x": np.array([px]),
                "particle_position_y": np.array([0.5]),
                "particle_position_z": np.array([0.5]),
                "particle_mass": np.array([1.0]),
            }

            ds = load_particles(particle_data)

            for dims in ([ncells] * 3, [ncells, ncells / 2, ncells / 4]):
                LE = np.array([0.05, 0.05, 0.05])
                RE = np.array([0.95, 0.95, 0.95])
                dims = np.array(dims)

                dds = (RE - LE) / dims
                volume = ds.quan(np.prod(dds), "cm**3")

                obj = ds.arbitrary_grid(LE, RE, dims)
                deposited_mass = obj[("deposit", "all_density")].sum() * volume

                assert_equal(deposited_mass, ds.quan(1.0, "g"))

                LE = np.array([0.00, 0.00, 0.00])
                RE = np.array([0.05, 0.05, 0.05])

                obj = ds.arbitrary_grid(LE, RE, dims)

                deposited_mass = obj[("deposit", "all_density")].sum()

                assert_equal(deposited_mass, 0)

    # Test that we get identical results to the covering grid for unigrid data.
    # Testing AMR data is much harder.
    for nprocs in [1, 2, 4, 8]:
        ds = fake_random_ds(32, nprocs=nprocs)
        for ref_level in [0, 1, 2]:
            cg = ds.covering_grid(
                ref_level, [0.0, 0.0, 0.0], 2**ref_level * ds.domain_dimensions
            )
            ag = ds.arbitrary_grid(
                [0.0, 0.0, 0.0], [1.0, 1.0, 1.0], 2**ref_level * ds.domain_dimensions
            )
            assert_almost_equal(cg[("gas", "density")], ag[("gas", "density")])


def test_octree_cg():
    ds = fake_octree_ds(num_zones=1, partial_coverage=0)
    cgrid = ds.covering_grid(
        0, left_edge=ds.domain_left_edge, dims=ds.domain_dimensions
    )
    density_field = cgrid[("gas", "density")]
    assert_equal((density_field == 0.0).sum(), 0)


def test_smoothed_covering_grid_2d_dataset():
    ds = fake_random_ds([32, 32, 1], nprocs=4)
    ds.force_periodicity()
    scg = ds.smoothed_covering_grid(1, [0.0, 0.0, 0.0], [32, 32, 1])
    assert_equal(scg[("gas", "density")].shape, [32, 32, 1])


def test_arbitrary_grid_derived_field():
    def custom_metal_density(field, data):
        # Calculating some random value
        return data["gas", "density"] * np.random.random_sample()

    ds = fake_random_ds(64, nprocs=8, particles=16**2)
    ds.add_field(
        ("gas", "Metal_Density"),
        units="g/cm**3",
        function=custom_metal_density,
        sampling_type="cell",
    )

    def _tracerf(field, data):
        return data[("gas", "Metal_Density")] / data["gas", "density"]

    ds.add_field(
        ("gas", "tracerf"),
        function=_tracerf,
        units="dimensionless",
        sampling_type="cell",
        take_log=False,
    )

    galgas = ds.arbitrary_grid([0.4, 0.4, 0.4], [0.99, 0.99, 0.99], dims=[32, 32, 32])
    galgas[("gas", "tracerf")]


def test_arbitrary_field_parameters():
    def _test_field(field, data):
        par = data.get_field_parameter("test_parameter")
        return par * data["all", "particle_mass"]

    ds = fake_random_ds(64, nprocs=8, particles=16**2)
    ds.add_field(
        ("all", "test_field"),
        units="g",
        function=_test_field,
        sampling_type="particle",
        validators=[ValidateParameter("test_parameter")],
    )

    agrid = ds.arbitrary_grid([0.4, 0.4, 0.4], [0.99, 0.99, 0.99], dims=[32, 32, 32])
    agrid.set_field_parameter("test_parameter", 2)
    assert_array_equal(2 * agrid["all", "particle_mass"], agrid["all", "test_field"])


def test_arbitrary_grid_edge():
    # Tests bug fix for issue #2087
    # Regardless of how left_edge and right_edge are passed, the result should be
    # a YTArray with a unit registry that matches that of the dataset.
    dims = [32, 32, 32]
    ds = fake_random_ds(dims)
    # Test when edge is a list, numpy array, YTArray with dataset units, and
    # YTArray with non-dataset units
    ledge = [
        [0.0, 0.0, 0.0],
        np.array([0.0, 0.0, 0.0]),
        [0.0, 0.0, 0.0] * ds.length_unit,
        [0.0, 0.0, 0.0] * kpc,
    ]

    redge = [
        [1.0, 1.0, 1.0],
        np.array([1.0, 1.0, 1.0]),
        [1.0, 1.0, 1.0] * ds.length_unit,
        [1.0, 1.0, 1.0] * kpc,
    ]

    ledge_ans = [
        [0.0, 0.0, 0.0] * ds.length_unit.to("code_length"),
        np.array([0.0, 0.0, 0.0]) * ds.length_unit.to("code_length"),
        [0.0, 0.0, 0.0] * ds.length_unit,
        [0.0, 0.0, 0.0] * kpc,
    ]

    redge_ans = [
        [1.0, 1.0, 1.0] * ds.length_unit.to("code_length"),
        np.array([1.0, 1.0, 1.0]) * ds.length_unit.to("code_length"),
        [1.0, 1.0, 1.0] * ds.length_unit,
        [1.0, 1.0, 1.0] * kpc,
    ]

    for le, re, le_ans, re_ans in zip(ledge, redge, ledge_ans, redge_ans):
        ag = ds.arbitrary_grid(left_edge=le, right_edge=re, dims=dims)
        assert np.array_equal(ag.left_edge, le_ans)
        assert np.array_equal(ag.right_edge, re_ans)
        assert ag.left_edge.units.registry == ds.unit_registry
        assert ag.right_edge.units.registry == ds.unit_registry
        ag[("gas", "density")]



import yt

# Load the dataset.
ds = yt.load("Enzo_64/DD0043/data0043")

# Create a volume rendering, which will determine data bounds, use the first
# acceptable field in the field_list, and set up a default transfer function.

# This will save a file named 'data0043_Render_density.png' to disk.
im, sc = yt.volume_render(ds, field=("gas", "density"))
